import io
import json
import os
import platform
import requests
import concurrent.futures
import threading
import time
import logging
from urllib.parse import urlparse

from typing import Optional

from arthurai.common.exceptions import MissingParameterError, InternalValueError
from arthurai.version import __version__

logger = logging.getLogger(__name__)


class BaseApiClient(object):
    """docstring for BaseApiClient"""

    def __init__(self, access_key=None, url=None, base_path=None, thread_workers=10, verify_ssl=True, offline=False, **kwargs):
        """Client which maintains REST calls and connection to API

        :param access_key: API key
        :param url: URL of the api/arthur server
        :param thread_workers: the number of workers the client requires when uploading inferences
        """
        self.access_key = kwargs.get('access_key', access_key)
        url = kwargs.get('url', url)
        if self.access_key is None:
            self.access_key = os.getenv('ARTHUR_API_KEY')
        if url is None:
            url = os.getenv('ARTHUR_ENDPOINT_URL')

        if url is None or self.access_key is None:
            raise MissingParameterError("Please set api key and url either via environment variables "
                                        "(`ARTHUR_API_KEY` and `ARTHUR_ENDPOINT_URL`) or by passing parameters "
                                        "`access_key` and `url`.")

        self.pool = concurrent.futures.ThreadPoolExecutor(max_workers=thread_workers)
        self.base_path = base_path
        self.verify_ssl = verify_ssl
        self.user_agent = None

        parsed_url = urlparse(url)
        if parsed_url.netloc and parsed_url.path:
            logger.warning(f"Path of '{parsed_url.path}' is not needed and is omitted.") 
        if parsed_url.scheme and not parsed_url.netloc:
            raise InternalValueError("Please make sure your url has a valid netloc.")
        if parsed_url.query:
            logger.warning(f"Query params of '{parsed_url.query}' is not needed and is omitted.") 

        if parsed_url.scheme in ('http', 'https'):
            self.url = parsed_url.scheme + "://" + (parsed_url.netloc or parsed_url.path)
        else:
            logger.warning(f"Url scheme of '{parsed_url.scheme}' is incorrect or not provided. Defaulted to 'https'.") 
            self.url = "https://" + (parsed_url.netloc or parsed_url.path)

        if not offline:
            user_org = self._get_current_user_org()
            agent_info = f"arthur-sdk/{__version__} (system={platform.system()}, org={user_org})" if user_org \
                else f"arthur-sdk/{__version__} (system={platform.system()})"
            self.user_agent = agent_info

    def _get_current_user_org(self) -> Optional[str]:
        url = f"{self.base_path}/users/me"
        resp = self.get(url, return_raw_response=True)
        if resp.status_code == 200:
            response_body = resp.json()
            if "organization_id" in response_body:
                return response_body["organization_id"]
        else:
            logger.debug(f"Failed to get current user organization, received response code {resp.status_code} with body {resp.content}")
        return None

    def send(self, url, method='GET', data=None, files=None, params=None, headers=None, return_raw_response=False,
             retries=0):
        """Sends the specified data with headers to the given url with the given request type

        :param retries: Number of times to retry the request if it results in a 400 or higher response code
        :param return_raw_response: If true do not filter request response, return raw object
        :param url: url to send data to
        :param method: REST call type, POST, GET, PUT, DELETE
        :param data: the data to send
        :param headers: headers to use in the REST call
        :return: response of the REST call
        """
        if headers is None:
            headers = {}

        if self.user_agent is not None:
            headers['User-Agent'] = self.user_agent

        if not url.startswith(('http://', 'https://')):
            url = '{base}{endpoint}'.format(
                base=self.url,
                endpoint=url
            )
        else:
            url = url

        # append the authentication headers to all requests
        headers = headers.copy()
        if self.access_key:
            if self.base_path and 'v3' in self.base_path:
                headers['Authorization'] = self.access_key
            else:
                headers['X-API-KEY'] = self.access_key

        headers['Content-Type'] = headers.get('Content-Type', 'application/json')
        headers['Accept'] = headers.get('Accept', 'application/json')

        # convert JSON data to a string
        if headers.get('Content-Type') == 'multipart/form-data':
            pass

        elif data:
            data = json.dumps(data)

        # send request to the test client and return the response
        if headers.get('Content-Type') == 'multipart/form-data':
            headers.pop('Content-Type')
            multipart = True
            logger.debug("Sending multipart request: %s %s\nHeaders: %s", method, url, headers)
            rv = requests.request(method, url, data=data, files=files, headers=headers, verify=self.verify_ssl)
        else:
            multipart = False
            logger.debug("Sending request: %s %s\nHeaders: %s\nData: %s", method, url, headers, data)
            rv = requests.request(method, url, data=data, params=params, headers=headers, verify=self.verify_ssl)

        attempt_retries = 0
        while attempt_retries < retries and rv.status_code >= 400:
            time.sleep(0.05)
            logger.debug(f"Request failed with status {rv.status_code} auto retry {attempt_retries + 1}/{retries}")
            if multipart:
                for fkey in files:
                    try:
                        files[fkey].seek(0)
                    except AttributeError:
                        continue
                rv = requests.request(method, url, data=data, files=files, headers=headers, verify=self.verify_ssl)
            else:
                rv = requests.request(method, url, data=data, params=params, headers=headers, verify=self.verify_ssl)
            attempt_retries += 1

        logger.debug(f"[{threading.current_thread().getName()}] Rest Call Response Time: {rv.elapsed.total_seconds() * 1000} ms")
        logger.debug("Received response: %d %s\nHeaders: %s\nContent: %s", rv.status_code, rv.url, rv.headers, rv.content)
        if return_raw_response:
            return rv

        return self._response(rv)

    def _response(self, rv):
        """Depending on the type of response from the server, parses the response and returns

        :param rv: response from the REST call
        :return: parsed response
        """
        # return error codes as raw responses
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
        if rv.status_code >= 400:
            return rv.content
        if rv.request.headers.get('Accept') == 'application/octet-stream':
            return io.BytesIO(rv.content)
        # return the content if the the return type is an image
        if rv.headers.get('Content-Type') == 'image/jpeg' \
                or rv.headers.get('Content-Type') == 'text/csv' \
                or rv.headers.get('Content-Type') == 'avro/binary' \
                or rv.headers.get('Content-Type') == 'parquet/binary':
            return rv.content

        try:
            response = rv.json()
        except ValueError:
            response = rv.content

        return response

    def get(self, url, headers={}, params=None, return_raw_response=False, retries=0):
        """
        Sends a GET request to the given url with the given headers
        :param retries: Number of times to retry the request if it results in a 400 or higher response code
        :param return_raw_response: If true do not filter response, return raw object
        :param url: url to send request to
        :param headers: headers to use in the request
        :return: response of the rest call
        """
        return self.send(url, 'GET', headers=headers, params=params, return_raw_response=return_raw_response,
                         retries=retries)

    def post(self, url, data, files=None, headers={}, return_raw_response=False, retries=0):
        """Sends a POST request to the given url with the given headers

        :param return_raw_response: If true do not filter response, return raw object
        :param url: url to send request to
        :param headers: headers to use in the request
        :return: response of the rest call
        :param retries: Number of times to retry the request if it results in a 400 or higher response code
        """
        return self.send(url, 'POST', data, files=files, headers=headers, return_raw_response=return_raw_response,
                         retries=retries)

    def patch(self, url, data, files=None, headers={}, return_raw_response=False, retries=0):
        """Sends a PATCH request to the given url with the given headers

        :param return_raw_response: If true do not filter response, return raw object
        :param url: url to send request to
        :param headers: headers to use in the request
        :param retries: Number of times to retry the request if it results in a 400 or higher response code
        :return: response of the rest call
        """
        return self.send(url, 'PATCH', data, files=files, headers=headers, return_raw_response=return_raw_response,
                         retries=retries)

    def put(self, url, data, headers={}, return_raw_response=False, retries=0):
        """
        Sends a PUT request to the given url with the given headers
        :param return_raw_response: If true do not filter response, return raw object
        :param url: url to send request to
        :param headers: headers to use in the request
        :param retries: Number of times to retry the request if it results in a 400 or higher response code
        :return: response of the rest call
        """
        return self.send(url, 'PUT', data, headers=headers, return_raw_response=return_raw_response,
                         retries=retries)

    def delete(self, url, headers={}, return_raw_response=False, retries=0):
        """Sends a DELETE request to the given url with the given headers

        :param return_raw_response: If true do not filter response, return raw object
        :param url: url to send request to
        :param headers: headers to use in the request
        :param retries: Number of times to retry the request if it results in a 400 or higher response code
        :return: response of the rest call
        """
        return self.send(url, 'DELETE', headers=headers, return_raw_response=return_raw_response,
                         retries=retries)

    @staticmethod
    def async_call(rest_call, pool, *args, callback=None):
        """Starts a new process asynchronously

        :param rest_call: a pointer to the rest call which should be executed async
        :param pool: python process pool to take processes from
        :param callback: function which will get called given the response of the child thread
        :return: returns a python AsyncResult object
        """
        logger.debug(f"Current pool queue size: {pool._work_queue.qsize()}")

        future = pool.submit(rest_call, *args)
        if callback is not None:
            future.add_done_callback(callback)
        return future



from datetime import datetime
import logging
from typing import Dict, List, Any, Optional, Union

from dateutil.parser import parse, ParserError
import pytz

from arthurai.common.exceptions import UserTypeError, UserValueError

logger = logging.getLogger(__name__)

"""
Only client facing utils should live in this funciton. If a util  is used for a specific package and not client facing
it should be added to that packages util.py file.
"""


def format_timestamps(inferences: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Checks list of json inferences to ensure their timestamps have been converted to ISO8601 conventions.

    :param inferences: Input to `send_inferences()` function.

    :return: updated list of json inferences with ISO8601 formatted timestamps
    :raise: TypeError: timestamp is not of type `datetime.datetime`
    :raise: ValueError: timestamp is not timezone aware and no location data is provided to remedy
    """
    # TODO: deprecate support for strings and delete conditional @ https://arthurai.atlassian.net/jira/software/projects/CR/boards/11?selectedIssue=CR-10
    if len(inferences) > 0 and (
        is_string_date(inferences[0].get("inference_timestamp"))
        or is_string_date(inferences[0].get("ground_truth_timestamp"))
    ):
        date_to_log = inferences[0].get("inference_timestamp") or inferences[0].get(
            "ground_truth_timestamp"
        )
        logger.warning(
            f"DEPRECATION WARNING: Your provided timestamps starting with: {date_to_log} is of type `str` and is assumed to be in UTC format. Timezone information will be ignored. `str` types will be deprecated. Please use timeaware `datetime.datetime` timestamps and use `arthurai.util.format_timestamp` if needed, e.g. `format_timestamp(datetime.strptime('2020-01-01 01:01:01', '%Y-%m-%d %H:%M:%S'), 'US/Eastern')`.",
        )

    for inf in inferences:
        for timestamp_key in ["inference_timestamp", "ground_truth_timestamp"]:
            if timestamp_key in inf:
                inf[timestamp_key] = format_timestamp(inf[timestamp_key])
    return inferences


def format_timestamp(
    timestamp: Union[datetime, str], location: Optional[str] = None
) -> str:
    """Check if timestamp is time aware and convert to ISO8601 conventioned string.

    This helper function converts datetime objects into timezone aware ISO8601 strings, which is necessary when sending JSON to Arthur's backend.
    If `timestamp` argument is timezone aware, no `location` needs to be provided; otherwise, a string pytz `location` like "US/Eastern"
    needs to be provided to establish timezone.
    String `timestamp` are supported for backwards compatability and for simplicity are assumed to already be in UTC format, but string support
    will be deprecated.

    :param timestamp: timestamp to format
    :param location: pytz location as string

    :return: ISO8601 formatted timestamp
    :raise: TypeError: timestamp is not of type `datetime.datetime`
    :raise: ValueError: timestamp is not timezone aware and no location data is provided to remedy
    """

    # TODO: deprecate support for strings and delete conditional @ https://arthurai.atlassian.net/jira/software/projects/CR/boards/11?selectedIssue=CR-10
    if isinstance(timestamp, str):
        if timestamp[-1] != "Z":
            return timestamp + "Z"
        return timestamp

    if not (location is None or isinstance(location, str)):
        raise UserTypeError(f"`location` must be valid pytz string like 'US/Eastern' or None type")

    localization = pytz.timezone(location) if location else None

    if not isinstance(timestamp, datetime):
        raise UserTypeError(
            f"`timestamp` {timestamp} must be of type `datetime.datetime`, not {type(timestamp)}."
        )

    if not (
        timestamp.tzinfo is not None
        and timestamp.tzinfo.utcoffset(timestamp) is not None
    ):
        if localization:
            return (
                localization.localize(timestamp).astimezone(pytz.utc).isoformat()[:-6]
                + "Z"
            )  # convert to utc, log warning if converting from non-utc to utc
        else:
            raise UserValueError(
                f"`timestamp` {timestamp} is not timezone aware, please supply pytz `timezone` string to make timestamp timezone aware, e.g. `format_timestamp(datetime.datetime(2021, 1, 1), 'US/Eastern')`."
            )
    return timestamp.astimezone(pytz.utc).isoformat()[:-6] + "Z"


# TODO: deprecate support for strings and delete method @ https://arthurai.atlassian.net/jira/software/projects/CR/boards/11?selectedIssue=CR-10
def is_string_date(string: Any, fuzzy=False) -> bool:
    """
    Return whether the string can be interpreted as a date.

    :param string: str, string to check for date
    :param fuzzy: bool, ignore unknown tokens in string if True
    """
    if not isinstance(string, str) or string.isdigit():
        return False
    try:
        parse(string, fuzzy=fuzzy)
    except ParserError:
        return False
    else:
        return True



from datetime import datetime, timezone

import responses
import pytest

from arthurai.common.constants import InputType, OutputType
from arthurai.core.models import ArthurModel
from tests.base_test import BaseTest
from arthurai import util as arthur_util


class TestInferences(BaseTest):

    def _mock_posting_inferences(self, model_id):
        server_response = {
            "counts": {
                "success": 1,
                "failure": 0,
                "total": 1
            },
            "failures": [
                {
                    "message": "Success",
                    "status": 200
                }
            ]
        }

        self.mockPost(f"/api/v3/models/{model_id}/inferences", server_response, status=207)

    def _create_mock_model(self):
        model_data = {
            "partner_model_id": "",
            "input_type": InputType.Tabular,
            "output_type": OutputType.Regression,
            "display_name": "",
            "description": "",
            "attributes": [],
        }
        model = ArthurModel(client=self.arthur.client, **model_data)
        model.id = "1234"
        return model

    def test_format_inference_no_gt(self):
        model = self._create_mock_model()

        inference = model._format_inference_request(
            inference_timestamp=datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
            partner_inference_id="1234",
            model_pipeline_input=ArthurModel._replace_nans_and_infinities_in_dict({"attr": 5}),
            non_input_data=ArthurModel._replace_nans_and_infinities_in_dict({"Gender": "male"}),
            predicted_value=ArthurModel._replace_nans_and_infinities_in_dict({"prediction": 800}),
            ground_truth=ArthurModel._replace_nans_and_infinities_in_dict(None)
        )

        inferences = arthur_util.format_timestamps([inference])
        assert len(inferences) == 1
        assert inferences[0]["inference_timestamp"] == "2020-08-13T17:44:31Z"
        assert "ground_truth_timestamp" not in inferences[0]
        assert "ground_truth" not in inferences[0]


    @responses.activate
    def test_format_inference_request(self):
        model = self._create_mock_model()

        inf = model._format_inference_request(
            inference_timestamp="2020-08-13T17:44:31.552125Z",
            partner_inference_id="1234",
            model_pipeline_input={"attr": 5},
            non_input_data={"Gender": "male"},
            predicted_value={"prediction": 800},
            ground_truth={"actual_value": 812}
        )

        expected = {
            "inference_timestamp": "2020-08-13T17:44:31.552125Z",
            "partner_inference_id": "1234",
            "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
            "ground_truth_timestamp": "2020-08-13T17:44:31.552125Z",
            "ground_truth_data": {"actual_value": 812}
        }

        assert inf == expected

    def test_update_inference_timestamp(self):
        inference_incorrect_timestamps = [
            {
                "inference_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "partner_inference_id": "1234",
                "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
                "ground_truth_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "ground_truth_data": {"actual_value": 812}
            },
            {
                "inference_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "partner_inference_id": "1234",
                "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
                "ground_truth_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "ground_truth_data": {"actual_value": 812}
            }
        ]

        reformatted_inferences = arthur_util.format_timestamps(inference_incorrect_timestamps)
        for inf in reformatted_inferences:
            assert inf["inference_timestamp"] == inf["ground_truth_timestamp"] == "2020-08-13T17:44:31Z"

    def test_string_update_inference_timestamp(self):
        inference_incorrect_timestamps = [
            {
                "inference_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "partner_inference_id": "1234",
                "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
                "ground_truth_timestamp": "2020-08-13T17:44:31.552125",
                "ground_truth_data": {"actual_value": 812}
            },
            {
                "inference_timestamp": "2020-08-13T17:44:31.552125",
                "partner_inference_id": "1234",
                "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
                "ground_truth_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "ground_truth_data": {"actual_value": 812}
            }
        ]

        reformatted_inferences = arthur_util.format_timestamps(inference_incorrect_timestamps)
        for inf in reformatted_inferences:
            assert inf["inference_timestamp"] in ["2020-08-13T17:44:31Z", "2020-08-13T17:44:31.552125Z"]
            assert inf["ground_truth_timestamp"] in ["2020-08-13T17:44:31Z", "2020-08-13T17:44:31.552125Z"]

        # TODO: uncomment this block and delete one above to depreciate support for strings
        # for obj in inference_incorrect_timestamps:
        #     with pytest.raises(TypeError):
        #         arthur_util.format_timestamps([obj])

    def test_format_timestamp_with_location(self):
        unaware_timestamp = datetime(2021, 2, 8, 12, 10, 5)
        assert arthur_util.format_timestamp(unaware_timestamp, "America/Phoenix") == "2021-02-08T19:10:05Z" 
        


from dataclasses import dataclass
from typing import Union, Optional

from arthurai.common.constants import ListableStrEnum
from arthurai.core.base import ArthurBaseJsonDataclass, NumberType


class AlertMetric(ListableStrEnum):
    TotalInferenceCount = "Total Inference Count"
    TotalInferenceCountByClass = "Total Inference Count By Class"
    AveragePrediction = "Average Prediction"
    RMSE = "RMSE"
    PSIDataDriftReferenceSet = "PSI Data Drift Reference Set"
    PSIBatchDataDriftReferenceSet = "PSI Batch Data Drift Reference Set"
    TruePositiveRate = "True Positive Rate"
    TrueNegativeRate = "True Negative Rate"
    FalsePositiveRate = "False Positive Rate"
    FalseNegativeRate = "False Negative Rate"


class AlertRuleBound(ListableStrEnum):
    Upper = "upper"
    Lower = "lower"


class AlertRuleSeverity(ListableStrEnum):
    Warning = "warning"
    Critical = "critical"


@dataclass
class AlertRule(ArthurBaseJsonDataclass):
    bound: AlertRuleBound
    threshold: NumberType
    metric: Union[AlertMetric, str]
    severity: AlertRuleSeverity
    query: Optional[dict] = None
    lookback_period: Optional[NumberType] = None
    subsequent_alert_wait_time: Optional[NumberType] = None
    check_batch_idle_time: Optional[NumberType] = None
    attribute_id: Optional[str] = None
    attribute_name: Optional[str] = None
    enabled: bool = True
    id: Optional[str] = None



import pytest

from arthurai.core.models import ArthurModel
from arthurai.core.alerts import AlertRule, AlertRuleBound, AlertMetric, AlertRuleSeverity
from arthurai.client.base import BaseApiClient
from tests import MockResponse


alert_rule_object = AlertRule(
    bound=AlertRuleBound.Upper,
    threshold=0.5,
    metric=AlertMetric.AveragePrediction,
    severity=AlertRuleSeverity.Warning
)

alert_rule_dict = {
    "bound": "upper",
    "threshold": 0.5,
    "metric": "Average Prediction",
    "severity": "warning",
    "enabled": True
}


def patch_alert_rule_post(*args, **kwargs):
    assert kwargs['data'] == alert_rule_dict
    alert_rule_object.id = "123456789"
    return MockResponse(alert_rule_object, 200)


def path_alert_rule_get(*args, **kwargs):
    alert_rule_list = {
      "data": [
        {
          "metric": "Total Inference Count By Class",
          "attribute_name": "dog",
          "threshold": 200,
          "bound": "lower",
          "severity": "warning",
          "lookback_period": 360,
          "subsequent_alert_wait_time": 720,
          "enabled": True,
          "id": "2905da23-fa72-4299-92d6-40007d3a2a03",
          "attribute_id": "37c8a2cb-b607-4b77-9a89-9d34ee6e3516"
        },
        {
          "metric": "Total Inference Count By Class",
          "attribute_name": "dog",
          "threshold": 200,
          "bound": "lower",
          "severity": "warning",
          "lookback_period": 360,
          "subsequent_alert_wait_time": 720,
          "enabled": True,
          "id": "2905da23-fa72-4299-92d6-40007d3a2a03",
          "attribute_id": "37c8a2cb-b607-4b77-9a89-9d34ee6e3516"
        }
      ],
      "page": 1,
      "page_size": 20,
      "total_pages": 1,
      "total_count": 20
    }
    return MockResponse(alert_rule_list, 200)


def patch_alert_rule_patch(*args, **kwargs):
    alert_rule_dict["id"] = "123456789"
    return MockResponse(alert_rule_dict, 200)


class TestAlertRule:

    def test_alert_rule_create(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        # test successful creation of alert rule
        monkeypatch.setattr(BaseApiClient, "post", patch_alert_rule_post)
        res = binary_classification_model.create_alert_rule(**alert_rule_dict)
        assert res == alert_rule_object

        # test creating alert rule using an AlertRule object
        res = binary_classification_model.create_alert_rule(alert_rule=alert_rule_object)
        assert res == alert_rule_object

    def test_get_alert_rules(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        monkeypatch.setattr(BaseApiClient, "get", path_alert_rule_get)
        res = binary_classification_model.get_model_alert_rules()
        assert len(res) == 2
        for alert_rule in res:
            assert isinstance(alert_rule, AlertRule)

    def test_alert_rule_patch(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        # test updating alert rule passing id separately
        monkeypatch.setattr(BaseApiClient, "patch", patch_alert_rule_patch)
        res = binary_classification_model.update_alert_rule(alert_rule_object, "123456789")
        alert_rule_object.id = "123456789"
        assert res == alert_rule_object

        # test creating alert rule using an AlertRule object
        res = binary_classification_model.update_alert_rule(alert_rule=alert_rule_object)
        alert_rule_object.id = "123456789"
        assert res == alert_rule_object

        # ensure an error is thrown if an alert rule id is not provided
        alert_rule_object.id = None
        with pytest.raises(Exception) as context:
            binary_classification_model.update_alert_rule(alert_rule_object)

        assert "alert_rule_to_update must have a valid id" in context.value.args[0]






from typing import List, Dict, Any, Optional, Tuple, TYPE_CHECKING
import shutil
import tempfile
import os
import importlib
import sys
import logging
import pkg_resources
import pandas as pd

from pathlib import PurePath
from zipfile import ZipFile

# need dill to be able to pickle LIME
import dill
import docker

from arthurai.common.constants import Stage, ValueType, InputType
from arthurai.common.exceptions import ExpectedParameterNotFoundError, ArthurUserError, InternalTypeError, \
    UserValueError
from arthurai.core.util import NumpyEncoder
from arthurai.explainability.arthur_explainer import ArthurExplainer
#  imports ArthurModel for type checking, required due to circular import
if TYPE_CHECKING:
    from arthurai.core.models import ArthurModel

logger = logging.getLogger(__name__)


class ExplanationPackager(object):
    """Manages the project source code and explainer for this model.
    Intended only for internal use. Used to create explainer object, package files and format config for the 
    update enrichments API endpoint.

    See model.update_enrichments() for more details on usage
    """

    # fields required to geenerate explainability files
    FILE_FIELDS = ['project_directory', 'user_predict_function_import_path', 'requirements_file', 'df', ]

    def __init__(self, model: 'ArthurModel', **kwargs):
        """Constructor for ExplanationPackager. Explanation packager needs to handle case of
        user setting files, and user just updating config. Because of this, don't raise any
        exceptions in constructor for missing required fields. Instead check and raise the exception in
        the function that needs the field.
        """
        self._model = model

        # use all kwargs so its easy to call constructor using config dict in model.update_enrichments
        self.df: pd.DataFrame = kwargs.get('df')
        self.project_directory: Optional[str] = kwargs.get("project_directory")
        self.streaming_explainability_enabled = kwargs.get("streaming_explainability_enabled")
        self.ignore_dirs = set(kwargs.get("ignore_dirs", []))
        self.user_predict_function_import_path: Optional[str] = kwargs.get("user_predict_function_import_path")
        self.requirements_file = kwargs.get("requirements_file")
        self.python_version = kwargs.get("python_version")
        self.sdk_version = kwargs.get("sdk_version")
        self.explanation_nsamples = kwargs.get("explanation_nsamples")
        self.explanation_algo = kwargs.get("explanation_algo")
        self.model_server_num_cpu = kwargs.get('model_server_num_cpu')
        self.model_server_memory = kwargs.get('model_server_memory')
        self.model_server_max_replicas = kwargs.get('model_server_max_replicas')
        self.inference_consumer_num_cpu = kwargs.get('inference_consumer_num_cpu')
        self.inference_consumer_memory = kwargs.get('inference_consumer_memory')
        self.inference_consumer_thread_pool_size = kwargs.get('inference_consumer_thread_pool_size')
        self.inference_consumer_score_percent = kwargs.get('inference_consumer_score_percent')
        self.text_delimiter = model.text_delimiter
        self.shap_nrows = kwargs.get("shap_nrows", 20)

        # explanation_algo is saved on the model object, and determines which explainer to use
        # however we always create both lime and shap, if possible, regardless of explanation_algo
        # this is to more easily allow switching between the 2
        self.enable_shap = self._model.input_type != InputType.NLP
        self.enable_lime = True
        if self._model.input_type == InputType.NLP and self.explanation_algo != ArthurExplainer.LIME:
            logger.warning("Only lime is supported for NLP explainabilty but shap was passed. "
                           "Automatically setting explanation_algo to lime")
            self.explanation_algo = ArthurExplainer.LIME

        self.explainer: ArthurExplainer = None
        self.zipfile: Optional[PurePath] = None

        self.dirname = "model_files"
        self.volume = os.path.join(tempfile.TemporaryDirectory(dir='/tmp').name, self.dirname)
        self.bind_path = "/mnt/model_files"
        self.test_script = "test.sh"
        self.example: Optional[List] = None

    def contains_file_fields(self) -> Tuple[bool, bool]:
        """Returns (onePresent, allPresent)
        onePresent is true if at least one file field was supplied
        AllPresent is true if all file fields are supplied
        """
        # NOTE if adding new file field check here, ensure its added to FILE_FIELDS at top of file
        checks = []
        checks.append(self.project_directory is not None)
        checks.append(self.user_predict_function_import_path is not None)
        checks.append(self.requirements_file is not None)
        checks.append(self.df is not None)

        return (any(checks), all(checks))

    def build_categorical_lookup_table(self) -> List[Optional[Dict[str, int]]]:
        """Builds the lookup table that ArthurExplainer needs for encoding strings"""
        num_attributes = len(self._model.get_attributes(Stage.ModelPipelineInput))
        label_mapping: List[Optional[Dict[str, int]]] = [None for _ in range(num_attributes)]
        sorted_attr = sorted(self._model.get_attributes(Stage.ModelPipelineInput), key=lambda x: x.position)
        for i, attr in enumerate(sorted_attr):
            if attr.categorical and attr.value_type == ValueType.String:
                label_mapping[i] = {val.value: j for j, val in enumerate(attr.categories)}
        return label_mapping

    def create(self) -> None:
        """Create an explainer for this model.

        :param: df: the training data for the explainer
        :param: shap_nrows: number of rows to sample the dataframe passed in
        """
        # this should never happen since we check for this at model onboarding
        if self._model.input_type == InputType.NLP and (self.text_delimiter is None or self.text_delimiter == ""):
            raise ExpectedParameterNotFoundError(f"Missing text delimiter. {InputType.NLP} require setting a text delimiter.")

        pipeline_attributes = self._model.get_attributes(Stage.ModelPipelineInput)
        pipeline_attributes_in_order = sorted(pipeline_attributes, key=lambda x: x.position)
        predicted_attrs = self._model.get_attributes(Stage.PredictedValue)

        # get categorical attr indices
        categorical_attr_indices = [i for i, attr in enumerate(pipeline_attributes_in_order) if attr.categorical]

        # pull out pipeline features from data
        attr_names = [attr.name for attr in pipeline_attributes_in_order]
        explanation_training_data = self.df[attr_names]
        self.example = list(explanation_training_data.iloc[0, :])

        if self.project_directory is None:
            raise InternalTypeError("project_directory is None")
        sys.path.insert(0, self.project_directory)
        current_dir = os.getcwd()
        os.chdir(self.project_directory)
        try:
            req_file_errors = self.validate_requirements_add_sklearn()
            if len(req_file_errors) > 0:
                error_str = '\n' + '\n'.join(req_file_errors)
                raise UserValueError(f"Failed to validate requirements file: {error_str}")

            if self.user_predict_function_import_path is None:
                raise InternalTypeError("user_predict_function_import_path is None")
            customer_model: Any = importlib.import_module(self.user_predict_function_import_path)
            label_mapping = self.build_categorical_lookup_table()
            self.explainer = ArthurExplainer(model_type=self._model.output_type,
                                             model_input_type=self._model.input_type,
                                             num_predicted_attributes=len(predicted_attrs),
                                             predict_func=customer_model.predict,
                                             data=explanation_training_data,
                                             categorical_features=categorical_attr_indices,
                                             label_mapping=label_mapping,
                                             shap_training_samples=self.shap_nrows,
                                             text_delimiter=self.text_delimiter,
                                             enable_shap=self.enable_shap)
        finally:
            sys.path.remove(self.project_directory)
            os.chdir(current_dir)

    def get_request_files(self) -> List[Any]:
        """Returns files needed for request, for use with model.update_enrichments
        """
        if self.zipfile is None:
            raise InternalTypeError("zipfile is None")
        if self.project_directory is None:
            raise InternalTypeError("project_directory is None")
        if self.requirements_file is None:
            raise InternalTypeError("requirements_file is None")
        return [
            ("user_project.zip", open(self.zipfile, "rb")),
            ("user_requirements_file.txt", open(os.path.join(self.project_directory, self.requirements_file))),
            ("explainer.pkl", dill.dumps(self.explainer)),
        ]

    def get_request_config(self) -> Dict[str, str]:
        """Generates config dict for explainer. For use with model.update_enrichments
        """
        config = {}

        # support shap not being enabled and not sending expected values
        if self.explainer is not None and self.explainer.enable_shap:
            expected_value = NumpyEncoder.convert_value(self.explainer.shap_expected_values())
            if not isinstance(expected_value, list):
                expected_value = [expected_value]
            config['shap_expected_values'] = expected_value

        if self.python_version:
            config['python_version'] = self.python_version.split()[0]
        if self.sdk_version:
            config['sdk_version'] = self.sdk_version.split()[0]
        if self.user_predict_function_import_path:
            config['user_predict_function_import_path'] = self.user_predict_function_import_path
        if self.explanation_algo:
            config['explanation_algo'] = self.explanation_algo
        if self.streaming_explainability_enabled is not None:
            config['streaming_explainability_enabled'] = self.streaming_explainability_enabled
        if self.explanation_nsamples:
            config['explanation_nsamples'] = self.explanation_nsamples
        if self.model_server_num_cpu:
            config['model_server_cpu'] = self.model_server_num_cpu
        if self.model_server_memory:
            config['model_server_memory'] = self.model_server_memory
        if self.model_server_max_replicas:
            config['model_server_max_replicas'] = self.model_server_max_replicas
        if self.inference_consumer_num_cpu:
            config['inference_consumer_cpu'] = self.inference_consumer_num_cpu
        if self.inference_consumer_memory:
            config['inference_consumer_memory'] = self.inference_consumer_memory
        if self.inference_consumer_thread_pool_size:
            config['inference_consumer_thread_pool_size'] = self.inference_consumer_thread_pool_size
        if self.inference_consumer_score_percent:
            config['inference_consumer_score_percent'] = self.inference_consumer_score_percent

        return config


    def make_zip(self) -> None:
        """Create the zip file for the project directory."""
        # ensure requirements file exists
        if self.project_directory is None:
            raise InternalTypeError("project_directory is None")
        if self.requirements_file is None:
            raise InternalTypeError("requirements_file is None")
        if not os.path.exists(os.path.join(self.project_directory, self.requirements_file)):
            raise ArthurUserError(
                "File {0} does not exist".format(os.path.join(self.project_directory, self.requirements_file)))

        zipfile = PurePath(self.volume, "project").with_suffix(".zip")
        base_dir = PurePath(self.project_directory)

        if not os.path.exists(self.volume):
            os.makedirs(self.volume)

        # create a ZipFile object
        with ZipFile(zipfile, 'w') as zip_file:
            # Iterate over all the files in directory
            for folder_name, sub_folders, filenames in os.walk(base_dir):
                for filename in filenames:
                    file_path = PurePath(folder_name, filename)
                    relative_path = file_path.relative_to(base_dir)
                    parents_set = set(map(str, relative_path.parents))
                    if parents_set.intersection(self.ignore_dirs):
                        print(f"Ignoring in zip file: {relative_path}")
                    else:
                        zip_file.write(file_path, relative_path)

        self.zipfile = zipfile

    def test(self) -> None:
        """Download the docker container, install the files, and make sure the predict function can be run."""
        example = NumpyEncoder.convert_value(self.example)
        self.copy_files(example)

        volumes = {self.volume: {"bind": self.bind_path}}
        client = docker.from_env()
        print("Building docker image...")
        image = client.images.build(path=os.path.dirname(__file__), pull=True, rm=True)[0]
        print("Creating container and testing predict function...")
        command = "{0}".format(os.path.join(self.bind_path, self.test_script))
        client.containers.run(image.id, volumes=volumes, command=command)
        print("Success!")

    def copy_files(self, example: Any) -> None:
        """Copy the files that are required for testing the docker image.

        :param: example: the example to use in the test script
        """

        explainer = os.path.join(self.volume, "explainer.pkl")
        dill.dump(self.explainer, open(explainer, 'wb'))

        if self.project_directory is None:
            raise InternalTypeError("project_directory is None")
        if self.requirements_file is None:
            raise InternalTypeError("requirements_file is None")
        requirements = os.path.join(self.volume, self.requirements_file)
        shutil.copy2(os.path.join(self.project_directory, self.requirements_file), requirements)

        test = open(os.path.join(self.volume, self.test_script), "w")
        test.write(self.create_test_script(example))
        os.chmod(os.path.join(self.volume, self.test_script), 0o755)

    def create_test_script(self, example: Any) -> str:
        """Create a test script.

        :param: example:the example to use in the test script

        :return: the script contents
        """
        script = ";".join([
            "from user_bundle.project import {0}".format(self.user_predict_function_import_path),
            "output = {0}.predict({1})".format(self.user_predict_function_import_path, example),
            "print(output)",
        ])

        return "\n".join([
            '#!/usr/bin/env bash',
            'set -e',
            'CODE_DIR="`pwd`/user_bundle"',
            'export USER_PROJECT_SOURCE_DIR="${CODE_DIR}/project"',
            'export USER_REQUIREMENTS_FILE="${CODE_DIR}/requirements.txt"',
            'export EXPLAINER_FILE="${CODE_DIR}/explainer.pkl"',
            'ENV_NAME="model-server"',
            'mkdir -p $CODE_DIR',
            'unzip -qq /mnt/{0}/project.zip -d $USER_PROJECT_SOURCE_DIR'.format(self.dirname),
            'cp /mnt/{0}/{1} $USER_REQUIREMENTS_FILE'.format(self.dirname, self.requirements_file),
            'cp /mnt/{0}/explainer.pkl $EXPLAINER_FILE'.format(self.dirname),
            'conda init bash',
            'conda create --name $ENV_NAME python=$PYTHON_VERSION --quiet',
            'source ~/.bashrc',
            'source activate $ENV_NAME',
            'pip -qq install --no-cache-dir -r $USER_REQUIREMENTS_FILE',
            'python -c "{0}"'.format(script),
        ])

    def validate_requirements_add_sklearn(self) -> List[str]:
        """Checks the user requirements file to make sure versions are correct.

        Also handles making sure scikit-learn version is pinned in user requirements
        file, if it is installed in local environment and not specified.

        scikit-learn is a dependency of shap, and arthurai doesn't require scikit-learn explicitly.
        This can lead to version A being installed and used locally when pickling explainer, then verion B
        getting installed on model server as part of shap install. This can lead to pickle errors,
        since the pickled version of shap (client side) used a different version of sklearn than what is on model server.
        """
        if self.requirements_file is None:
            raise InternalTypeError("requirements_file is None")
        with open(self.requirements_file, 'r') as f:
            user_reqs = f.readlines()

        # build dict of installed packages and versions
        cur_env = {
            pkg.project_name: pkg.version
            for pkg in pkg_resources.working_set
        }

        errors = []

        # validate each line in user requirements file
        sklearn_found = False
        for req in user_reqs:
            req = req.replace('\n', '')
            # validate single pinned version: pandas==1.0.0
            if '==' not in req:
                errors.append(f"Invalid line in requirements file: {req}. Each dependency must have a single specific version pinned using '=='")
                continue
            splt = req.split('==')
            if len(splt) != 2:
                errors.append(f"Invalid line in requirements file: {req}. Each dependency must have a single specific version pinned using '=='")

            # flag if sklearn is already pinned in requirements file
            package, version = splt
            sklearn_found = package == 'scikit-learn'

            # validate package version matches environment
            if package not in cur_env:
                logger.warn(f"Explainability requirements file contains package not installed in current environment: {package}.")
            if package in cur_env and cur_env[package] != version:
                error_str = f"Explainability requirements file lists {package}=={version} but found version {cur_env[package]} currently installed."
                error_str += " Packages listed in the requirements file must match the currently installed version."
                error_str += f" To resolve, update the requirements file to version {package}=={cur_env[package]}"
                errors.append(error_str)

        # insert sklearn into requirements file if necessary
        if 'scikit-learn' in cur_env and not sklearn_found:
            cur_version = cur_env['scikit-learn']
            if self.requirements_file is None:
                raise InternalTypeError("requirements_file is None")
            with open(self.requirements_file, 'a') as f:
                # ensure we add on new line
                leading_newline = '' if user_reqs[-1][-1] == '\n' else '\n'
                f.write(f"{leading_newline}scikit-learn=={cur_version}\n")
            logger.info("scikit-learn was not included in explainer requirements file, but is installed in current environment. "
                        "Automatically adding current version to requirements file. This helps ensure correct environment on model server.")

        return errors



import pytest

from arthurai.common.exceptions import *


def test_user_value_error_is_arthur_user_error():
    try:
        raise UserValueError("bad value")
    except ArthurUserError:
        print("expected exception caught")


def test_user_value_error_is_arthur_error():
    try:
        raise UserValueError("bad value")
    except ArthurError:
        print("expected exception caught")


def test_user_value_error_is_value_error():
    try:
        raise UserValueError("bad value")
    except ValueError:
        print("expected exception caught")


def test_internal_value_error_is_arthur_internal_error():
    try:
        raise InternalValueError("bad value")
    except ArthurInternalError:
        print("expected exception caught")


def test_internal_value_error_is_arthur_error():
    try:
        raise InternalValueError("bad value")
    except ArthurError:
        print("expected exception caught")


def test_internal_value_error_is_value_error():
    try:
        raise InternalValueError("bad value")
    except ValueError:
        print("expected exception caught")


def test_arthur_internal_error_reraised():
    @arthur_excepted("expected failure")
    def internal_error():
        raise ArthurInternalError

    with pytest.raises(ArthurInternalError):
        internal_error()


def test_arthur_user_error_reraised():
    @arthur_excepted("expected failure")
    def user_error():
        raise ArthurUserError

    with pytest.raises(ArthurUserError):
        user_error()


def test_arthur_external_error():
    @arthur_excepted("expected failure")
    def external_error():
        raise ValueError

    with pytest.raises(ArthurInternalError):
        external_error()



import pytest
import mock
import os

from arthurai.core.models import ArthurModel
from arthurai.client.base import BaseApiClient
from tests.base_test import BaseTest


@pytest.fixture(scope="session")
def mock_cred_env_vars():
    with mock.patch.dict(os.environ, {"ARTHUR_API_KEY": "access_key", "ARTHUR_ENDPOINT_URL": BaseTest._base_url}):
        yield


@pytest.fixture(scope="session")
def binary_classification_model() -> ArthurModel:
    binary_classification_model = {
        "id": "4a57c553-e787-4307-88f1-3747ec9130f5",
        "partner_model_id": "binary_classification_model",
        "input_type": "TABULAR",
        "output_type": "MULTICLASS",
        "archived": False,
        "created_at": "2020-10-26T21:43:24.632335Z",
        "updated_at": "2020-10-26T21:43:26.364596Z",
        "attributes": [
            {
                "id": "1d4bec8d-84a6-4ae7-a7ea-1a2230bb977c",
                "name": "input",
                "value_type": "INTEGER",
                "stage": "PIPELINE_INPUT",
                "position": 1,
                "categorical": False,
                "monitor_for_bias": False,
                "is_positive_predicted_attribute": False
            },
            {
                "id": "254c8f5f-f4b8-4943-89bf-db106d5ca835",
                "name": "pos_predicted_probability",
                "value_type": "FLOAT",
                "stage": "PREDICTED_VALUE",
                "position": 1,
                "categorical": False,
                "min_range": 0,
                "max_range": 1,
                "monitor_for_bias": False,
                "is_unique": False,
                "is_positive_predicted_attribute": True,
                "attribute_link": "pos_gt"
            },
            {
                "id": "ac97a47f-f755-451e-956c-d1ed52403915",
                "name": "pos_gt",
                "value_type": "INTEGER",
                "stage": "GROUND_TRUTH",
                "position": 1,
                "categorical": False,
                "min_range": 0,
                "max_range": 1,
                "monitor_for_bias": False,
                "is_unique": True,
                "is_positive_predicted_attribute": False,
                "attribute_link": "pos_predicted_probability"
            },
            {
                "id": "c2e626df-ee4c-403f-818b-f4f2d9180eab",
                "name": "neg_gt",
                "value_type": "INTEGER",
                "stage": "GROUND_TRUTH",
                "position": 2,
                "categorical": False,
                "min_range": 0,
                "max_range": 1,
                "monitor_for_bias": False,
                "is_unique": True,
                "is_positive_predicted_attribute": False,
                "attribute_link": "neg_predicted_probability"
            },
            {
                "id": "d4f04345-c92c-4959-9d92-7fe35924efeb",
                "name": "neg_predicted_probability",
                "value_type": "FLOAT",
                "stage": "PREDICTED_VALUE",
                "position": 2,
                "categorical": False,
                "min_range": 0,
                "max_range": 1,
                "monitor_for_bias": False,
                "is_unique": False,
                "is_positive_predicted_attribute": False,
                "attribute_link": "neg_gt"
            }
        ],
        "tags": [],
        "classifier_threshold": 0.5,
        "explainability": {
            "enabled": False
        },
        "is_batch": False,
        "text_delimiter": ","
    }
    yield ArthurModel(client=BaseApiClient(url="test.ai", offline=True), **binary_classification_model)



import json
import logging
import os
from pathlib import Path
import re
from typing import List, Union

import numpy as np
import pandas as pd
import requests

import arthurai.util as arthur_util
from arthurai.common.exceptions import InternalValueError, UserTypeError, UserValueError, InternalTypeError

logger = logging.getLogger(__name__)

def retrieve_parquet_files(directory_path: str):
    """Checks whether a given directory and its subdirectories contain parquet files,
    if so this will return a list of the files

    :param directory_path:    local path to check files types

    :return: List of paths for parquet files that are found
    """
    parquet_files = []
    for path, subdir, files in os.walk(directory_path):
        for file in files:
            if file.endswith('.parquet'):
                parquet_files.append(Path(os.path.join(path, file)))
    return parquet_files


class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """

    @staticmethod
    def convert_value(obj):
        """Converts the given object from a numpy data type to a python data type, if the object is already a
        python data type it is returned

        :param obj: object to convert
        :return: python data type version of the object
        """
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        elif isinstance(obj, pd.Timestamp):
            return obj.isoformat()
        elif isinstance(obj, float) and np.isnan(obj):
            return None
        return obj


def standardize_pd_obj(data: Union[pd.DataFrame, pd.Series], dropna: bool, replacedatetime: bool) -> Union[pd.DataFrame, pd.Series]:
    """Standardize pandas object for nans and datetimes.
    
    Standardization includes casting correct type for int columns that are float due to nans and 
    for converting datetime objects into isoformatted strings.

    :param data: the pandas data to standardize
    :param dropna: if True, drop nans from numeric date columns
    :param dropna: if True, replace timestamps with isoformatted strings

    :return: the standardized pandas data
    :raise: TypeError: timestamp is not of type `datetime.datetime`
    :raise: ValueError: timestamp is not timezone aware and no location data is provided to remedy
    """
    
    def standardize_pd_series(series: pd.Series) -> pd.Series:
        series = series.replace([np.inf, -np.inf], np.nan)
        nans_present = series.isnull().values.any()
        if dropna:
            series = series.dropna()

        # handle case where int column has nans and therefore is incorrectly seen as float column
        if re.search("float*", series.dtype.name, flags=re.I) and nans_present:
            valid_series = series.dropna()
            if len(valid_series) > 0 and np.array_equal(valid_series, valid_series.astype(int)):
                return series.astype(pd.Int64Dtype())

        # check to make sure datetimes are timezone aware
        elif re.search("(datetime|timestamp).*", series.dtype.name, flags=re.I) or arthur_util.is_string_date(series.values[0]): # TODO: remove is_date and first conditional below when deprecating string support @ https://arthurai.atlassian.net/jira/software/projects/CR/boards/11?selectedIssue=CR-10
            if arthur_util.is_string_date(series.values[0]):
                logger.warning(
                    f"DEPRECATION WARNING: Your feature <{series.name}> providing timestamps starting with <{series.values[0]}> is of type `str` and is assumed to be in UTC format. `timezone` information will be ignored. `str` types will be deprecated. Please use timeaware `datetime.datetime` timestamps and use arthurai.util.format_timestamp if needed, e.g. `format_timestamp(datetime.strptime('2020-01-01 01:01:01', '%Y-%m-%d %H:%M:%S'), 'US/Eastern')`.",
                )
            formatted_series = series.apply(arthur_util.format_timestamp)
            if replacedatetime:
                return formatted_series

        return series

    if isinstance(data, pd.Series):
        return standardize_pd_series(data)
    elif isinstance(data, pd.DataFrame):
        if dropna:
            raise InternalValueError(f"Cannot use dropna={dropna} with data argument as pd.DataFrame.")
        df = data.copy()
        for column in df.columns:
            df[column] = standardize_pd_series(df[column])
        return df
    else:
        raise InternalTypeError("Cannot standardize object that is not pd.DataFrame or pd.Series.")


# STICK TO SEMANTIC VERSIONING:
# https://semver.org/
# TLDR:
# Given a version number MAJOR.MINOR.PATCH, increment the:
#
# MAJOR version when you make incompatible API changes,
# MINOR version when you add functionality in a backwards compatible manner, and
# PATCH version when you make backwards compatible bug fixes.
__version__ = "3.6.2"



import pandas as pd
import responses
from arthurai.common.constants import InputType, OutputType

from arthurai.core.models import ArthurModel
from tests.base_test import BaseTest


class TestReferenceData(BaseTest):

    def _mock_sending_reference_set(self, model_id):
        server_response = {
            "counts": {
                "success": 0,
                "failure": 1,
                "total": 1
            },
            "failures": [
                {
                    "message": "missing field",
                    "status": 400
                }
            ]
        }

        self.mockPost(f"/api/v3/models/{model_id}/reference_data", server_response, status=207)
        self.mockPost(f"/api/v3/models/{model_id}/reference_data", request_type=responses.PATCH, status=200,
                      dict_to_return={"message": "success"})

    @responses.activate
    def test_send_parquet_file(self):
        model_data = {
            "partner_model_id": "",
            "input_type": InputType.Tabular,
            "output_type": OutputType.Regression,
            "display_name": "",
            "description": "",
            "attributes": [],
        }
        model = ArthurModel(client=self.arthur.client, **model_data)
        model.id = "1234"
        self._mock_sending_reference_set("1234")
        resp, resp_close = model.set_reference_data(data=pd.DataFrame({"col": [1, 2, 3, 4]}))
        assert resp["counts"]["failure"] == 1
        assert resp_close is None



import datetime
import uuid
from arthurai.common.constants import Stage, OutputType
from arthurai.common.exceptions import ExpectedParameterNotFoundError, MethodNotApplicableError
from arthurai.core.models import ArthurModel
import pandas as pd
from typing import Optional, Any, Tuple, Callable
import pytz



def log_prediction(arthur_model: ArthurModel) -> Callable[[Any], Any]:
    """Decorator to log the inputs and prediction of a model to Arthur

    :param arthur_model: A previously-saved ArthurModel object

    Note, the prediction function to be decorated can take optional arguments for logging.
    :param timestamp: A timestamp in ISO 8601 format
    :param partner_inference_id: A unique identifier for an inference

    :return: Tuple of (model_prediction, inference_id)
    """
    if arthur_model.output_type not in [OutputType.Regression,  OutputType.Multiclass]:
        raise MethodNotApplicableError("Logging decorator supports only Regression and Multiclass output types.")

    def decorator_send(func):
        def wrapper_send(*args, **kwargs):
            try:
              timestamp = kwargs.pop("inference_timestamp")
            except KeyError:
              timestamp = datetime.datetime.now(pytz.utc)

            try:
              inference_id = kwargs.pop("partner_inference_id")
            except KeyError:
              inference_id = str(uuid.uuid4())

            output = func(*args)

            inference_data = _format_input_data(*args, arthur_model)
            inference_data.update(
                _format_predicted_value_data(output, arthur_model)
            )

            arthur_model.send_inferences([{
                "inference_data": inference_data,
                "partner_inference_id": inference_id,
                "inference_timestamp": timestamp
            }])

            return output, inference_id
        return wrapper_send
    return decorator_send


def _get_attr_ordering(arthur_model):
    """ Return attribute names in order sorted by attr.position
    """
    model_input_attrs = [(attr.name, attr.position) \
                         for attr in arthur_model.get_attributes(Stage.ModelPipelineInput)]
    model_input_attrs.sort(key=lambda t: t[1])
    input_attribute_names = [name for (name, pos) in model_input_attrs]

    predicted_val_attrs = [(attr.name, attr.position) \
                           for attr in arthur_model.get_attributes(Stage.PredictedValue)]
    predicted_val_attrs.sort(key=lambda t: t[1])
    predicted_val_attribute_names = [name for (name, pos) in predicted_val_attrs]
    return input_attribute_names, predicted_val_attribute_names


def _format_input_data(input_vec, arthur_model):
    """ Arrange in ModelPipelineInput data into proper order and construct dict.
    """
    if isinstance(input_vec, pd.DataFrame):
        return input_vec.to_dict(orient="rows")[0]
    else:
        raise MethodNotApplicableError(str(input_vec.__class__) + " not yet supported.")
        input_attribute_names, predicted_val_attribute_names = _get_attr_ordering(arthur_model)
        return dict(zip(input_vec, input_attribute_names))

def _format_predicted_value_data(model_output, arthur_model):
    """ Arrange PredictedValue data into proper order and construct dict.
    """
    try:
        # wrap in an iterable so the model output can be zipped with column names
        # primarily relevant for regression models with a single output
        model_output.__iter__()
    except AttributeError:
        model_output = [model_output]

    _validate_predicted_value_attrs(arthur_model)
    input_attribute_names, predicted_val_attribute_names = _get_attr_ordering(arthur_model)
    return dict(zip(predicted_val_attribute_names, model_output))

def _validate_predicted_value_attrs(arthur_model):
    """Check if the positions of the PredictedValue attributes have been set. If any are None
    then we can't complete the request.
    """
    attr_positions = [attr.position for attr in arthur_model.get_attributes(Stage.PredictedValue)]
    if any([p is None for p in attr_positions]):
        raise ExpectedParameterNotFoundError("Unable to send inference. All PredictedValue attributes must have `position` assigned.")



from arthurai.explainability.explanation_packager import ExplanationPackager
from arthurai.core.models import ArthurModel
from arthurai.common.constants import InputType, OutputType

import pkg_resources


def test_validate_requirements() -> None:
    # build dict of installed packages and versions
    
    # gather user env, skipping sklearn, since we don't want to grab that when we build test req file
    cur_env = [
        (pkg.project_name, pkg.version)
        for pkg in pkg_resources.working_set
        if pkg.project_name != 'scikit-learn'
    ]

    # generate a requirements file that should succeed
    req_path = '/tmp/user_requirements.txt'
    with open(req_path, 'w') as f:
        test_package, test_version = cur_env[0]
        f.write(f"{test_package}=={test_version}")
    
    model = ArthurModel("123", InputType.Tabular, OutputType.Regression)
    packager = ExplanationPackager(model, requirements_file=req_path)

    # validate no errors
    errors = packager.validate_requirements_add_sklearn()
    assert len(errors) == 0

    # validate sklearn was added
    # we should always have sklearn in testing environment, since bunlded with shap
    with open(req_path, 'r') as f:
        req_lines = f.readlines()

    assert len(req_lines) == 2
    assert 'scikit-learn' in req_lines[1]

    # change version in req file to fail
    with open(req_path, 'w+') as f:
        test_package, test_version = cur_env[0]
        test_version = "-1.-1.-1"
        f.write(f"{test_package}=={test_version}")

    packager = ExplanationPackager(model, requirements_file=req_path)

    # validate no errors
    errors = packager.validate_requirements_add_sklearn()
    assert len(errors) == 1
    



import logging
import os
import tempfile
from typing import Dict, Optional, Any, TYPE_CHECKING, List

from pandas import DataFrame

from arthurai.common.constants import Stage
from arthurai.common.exceptions import UserValueError
from arthurai.core import util
#  imports ArthurModel for type checking, required due to circular import
if TYPE_CHECKING:
    from arthurai.core.models import ArthurModel

logger = logging.getLogger(__name__)


class DatasetService:
    COUNTS = "counts"
    SUCCESS = "success"
    FAILURE = "failure"
    TOTAL = "total"
    FAILURES = "failures"

    @staticmethod
    def convert_dataframe(model_id: str, stage: Optional[Stage], df: DataFrame) -> str:
        """Convert a dataframe to parquet named {model.id}-{stage}.parquet in the system tempdir

        :param model_id: a model id
        :param stage: the :py:class:`.Stage`
        :param df: the dataframe to convert

        Returns:
            The filename of the parquet file that was created
        """
        name = "{0}-{1}.parquet".format(model_id, stage) if stage else "{0}.parquet".format(model_id)
        filename = os.path.join(tempfile.mkdtemp(), name)
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass

        df.to_parquet(filename, index=False, allow_truncated_timestamps=True)
        return filename

    @staticmethod
    def send_parquet_files_from_dir_iteratively(model: 'ArthurModel', directory_path: str,
                                                url: str, upload_file_param_name: str,
                                                additional_form_params: Optional[Dict[str, Any]] = None,
                                                retries: int = 0):
        """Sends parquet files iteratively from a specified directory to a specified url for a given model

        :param retries: Number of times to retry the request if it results in a 400 or higher response code
        :param model:    the :py:class:`!arthurai.client.apiv2.model.ArthurModel`
        :param directory_path:    local path containing parquet files to send
        :param url:    POST url endpoint to send files to
        :param upload_file_param_name:     name to use in body with attached files
        :param additional_form_params: dictionary of additional form file params to send along with parquet file

        :raises MissingParameterError: the request failed

        :returns A list of files which failed to upload
        """
        files = util.retrieve_parquet_files(directory_path)
        if not files:
            raise UserValueError("The directory supplied does not contain any parquet files to upload")

        failed_files = []
        succeeded_files = []
        expected_keys = {DatasetService.SUCCESS, DatasetService.FAILURE, DatasetService.TOTAL}

        counts = {
            DatasetService.SUCCESS: 0,
            DatasetService.FAILURE: 0,
            DatasetService.TOTAL: 0
        }
        failures: List[Any] = []

        for file in files:
            if file.suffix == '.parquet':
                with open(file, 'rb') as parquet_file:
                    headers = {'Content-Type': 'multipart/form-data'}
                    form_parts = {} if additional_form_params is None else additional_form_params
                    form_parts.update({upload_file_param_name: parquet_file})
                    resp = model._client.post(url, data=None, files=form_parts, headers=headers,
                                              return_raw_response=True, retries=retries)
                    if resp.status_code == 201:
                        logger.info(f"Uploaded completed: {file}")
                        succeeded_files.append(file)
                    elif resp.status_code == 207:
                        logger.info(f"Upload completed: {file}")
                        result: Dict[str, Dict[str, int]] = resp.json()
                        # ensure the response is in the correct format
                        if DatasetService.COUNTS in result and DatasetService.FAILURES in result \
                                and set(result[DatasetService.COUNTS].keys()) == expected_keys:
                            counts[DatasetService.SUCCESS] += \
                                result[DatasetService.COUNTS][DatasetService.SUCCESS]
                            counts[DatasetService.FAILURE] += \
                                result[DatasetService.COUNTS][DatasetService.FAILURE]
                            counts[DatasetService.TOTAL] += \
                                result[DatasetService.COUNTS][DatasetService.TOTAL]
                            failures.append(result[DatasetService.FAILURES])
                        else:
                            failures.append(result)
                    else:
                        logger.error(f"Failed to upload file: {resp.text}")
                        failed_files.append(file)
                        failures.append(resp.json())
                        counts[DatasetService.FAILURE] += 1
                        counts[DatasetService.TOTAL] += 1

        file_upload_info = {
            DatasetService.COUNTS: counts,
            DatasetService.FAILURES: failures
        }

        # Only log failed or succeeded files if they exist
        if len(failed_files) > 0:
            logger.error(f'Failed to upload {len(failed_files)} files')
        if len(succeeded_files) > 0:
            logger.info(f'Successfully uploaded {len(succeeded_files)} files')
        return failed_files, file_upload_info



import logging
import re
import sys
from datetime import datetime
from http import HTTPStatus
from typing import Optional, Union, List, Dict, Any
import json

import numpy as np
import pyarrow.parquet as pq
from dataclasses import dataclass, InitVar
from pandas import DataFrame, Series

from arthurai.version import __version__
from arthurai import util as arthur_util
from arthurai.core import util as core_util
from arthurai.client.base import BaseApiClient
from arthurai.client import validation
from arthurai.common.constants import Stage, InputType, OutputType, ValueType, TextDelimiter, Enrichment
from arthurai.common.exceptions import MethodNotApplicableError, MissingParameterError, UserValueError, \
    UserTypeError, ArthurUserError, ExpectedParameterNotFoundError, arthur_excepted, InternalTypeError
from arthurai.core.base import ArthurBaseJsonDataclass, NumberType
from arthurai.core.attributes import ArthurAttribute, AttributeCategory, AttributeBin
from arthurai.core.data_service import DatasetService
from arthurai.explainability.explanation_packager import ExplanationPackager
from arthurai.core.alerts import AlertRule, AlertRuleSeverity, AlertRuleBound, AlertMetric
from arthurai.core.viz.visualizer import DataVisualizer

logger = logging.getLogger(__name__)

INFERENCE_DATA_RETRIES = 3


@dataclass
class ExplainabilityParameters(ArthurBaseJsonDataclass):
    enabled: bool
    explanation_algo: Optional[str] = None
    model_server_cpu: Optional[str] = None
    model_server_memory: Optional[str] = None
    explanation_nsamples: Optional[int] = None


@dataclass
class ArthurModel(ArthurBaseJsonDataclass):
    """
    Arthur Model class represents the metadata which represents a registered model in the application

    :param client: :class:`arthurai.client.Client` object which manages data storage
    :param partner_model_id: Client provided unique id to associate with the model. This field must be unique across
                             all active models cannot be changed once set.
    :param input_type: :class:`arthurai.common.constants.InputType` representing the model's input data type.
    :param output_type: :class:`arthurai.common.constants.InputType` representing the model's output data format.
    :param explainability:  :class:`arthurai.core.models.ExplainabilityParameters` object representing parameters that
                            will be used to create inference explanations.
    :param id: The auto-generated unique UUID for the model. Will be overwritten if set by the user.
    :param display_name: An optional display name for the model.
    :param description: Optional description of the model.
    :param is_batch: Boolean value to determine whether the model sends inferences in batch or streaming format.
                     Defaults to False.
    :param archived: Indicates whether or not a model has been archived, defaults to false.
    :param created_at: UTC timestamp in ISO8601 format of when the model was created. Will be overwritten if set by the
                       user.
    :param updated_at: UTC timestamp in ISO8601 format of when the model was last updated. Will be overwritten if set by
                       the user.
    :param attributes: List of :class:`arthurai.core.attributes.ArthurAttribute` objects registered to the model
    :param tags: List of string keywords to associate with the model.
    :param classifier_thresholds: Threshold value for classification models, default is 0.5.
    :param text_delimiter: Only valid for models with input_type equal to
                           :py:attr:`arthurai.common.constants.InputType.NLP`. Represents the text delimiter
                           to divide input strings.
    :param expected_throughput_gb_per_day: Expected amount of throughput.
    """

    partner_model_id: str
    input_type: InputType
    output_type: OutputType
    # This is just used during init and will not be associated with
    # an instance of this class, instead reference ArthurModel._client
    client: InitVar[Optional[BaseApiClient]] = None
    explainability: Optional[ExplainabilityParameters] = None
    id: Optional[str] = None
    display_name: Optional[str] = None
    description: Optional[str] = None
    is_batch: bool = False
    archived: bool = False
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    attributes: Optional[List[ArthurAttribute]] = None
    tags: Optional[List[str]] = None
    classifier_threshold: Optional[float] = None
    text_delimiter: Optional[TextDelimiter] = None
    expected_throughput_gb_per_day: Optional[int] = None

    def __post_init__(self, client: BaseApiClient):
        # variables created here will only be accessible directly on the object itself, they will not be in the result
        # of object.to_dict() even if marked as public (does not have preceding underscore)
        self._client = client
        self._explainer: Optional[ExplanationPackager] = None
        self.viz = DataVisualizer(self)

    def __str__(self):
        return str(self.to_dict())

    @arthur_excepted("failed to save model")
    def save(self) -> str:
        """Check and save this model.  Sets and returns the ID on successful upload.

        :return: The model id

        :raise: Exception: the model has already been saved
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        # if the model has not been set yet
        if self.id is not None:
            raise MethodNotApplicableError("Cannot save a registered model; use update instead.")

        self._check_model_valid()

        if self.display_name is None:
            self.display_name = self.partner_model_id

        data = self.to_dict()
        resp = self._client.post(self._client.base_path + "/models", data=data, return_raw_response=True)
        validation.validate_response_status(resp, expected_status_code=HTTPStatus.CREATED)
        if 'id' not in resp.json():
            raise ExpectedParameterNotFoundError(f"An error occurred: {resp}, {resp.status_code}, {resp.content}")

        # update fields with response from API
        self.id = resp.json()['id']

        # TODO: if explainability is enabled send resources

        return self.id  # type: ignore

    def _check_model_valid(self) -> bool:
        """Check the validity of the model before saving, and prints out any errors it finds.

        (As time passes we can add more to this function to enforce more requirements.)

        :return: True if the model is valid, False otherwise.
        """

        # has attributes
        if self.attributes is None:
            # future enhancement, attributes should be initialized to empty not None
            raise InternalTypeError("attributes is None")
        if len(self.attributes) == 0:
            raise MissingParameterError("must add attributes to model before saving; see below for requirements.")

        # contains ground truth, input, and predicted attributes
        contains_gt = False
        contains_pred = False
        contains_ipt = False
        for attr in self.attributes:
            if attr.stage == Stage.GroundTruth:
                contains_gt = True
            if attr.stage == Stage.PredictedValue:
                contains_pred = True
            if attr.stage == Stage.ModelPipelineInput or attr.stage == Stage.PredictFunctionInput:
                contains_ipt = True
        if not contains_gt:
            raise MissingParameterError("does not contain any attribute with Stage.GroundTruth.")
        if not contains_pred:
            raise MissingParameterError("does not contain any attribute with Stage.PredictedValue.")
        if not contains_ipt:
            raise MissingParameterError("does not have any attribute with Stage.ModelPipelineInput or Stage.PredictFunctionInput.")

        # for binary models, need one predicted attribute to be positive
        predicted_value_attributes = self.get_attributes(stage=Stage.PredictedValue)
        if len(predicted_value_attributes) == 2 and self.output_type == OutputType.Multiclass:  # is binary model
            has_pos_pred = False
            for pred in predicted_value_attributes:
                if pred.is_positive_predicted_attribute:
                    has_pos_pred = True
            if not has_pos_pred:
                raise MethodNotApplicableError("binary models must have a positive predicted attribute; use "
                                               "add_binary_classifier_output_attributes instead.")

        return True

    @arthur_excepted("failed to update model")
    def update(self):
        """Update the the current model object 

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if self.id is None:
            raise MethodNotApplicableError("Model has not been created yet, use save() to instantiate this model object")

        data = self.to_dict()
        resp = self._client.put(self._client.base_path + f"/models/{self.id}", data=data, return_raw_response=True)
        validation.validate_response_status(resp, expected_status_code=HTTPStatus.OK)

    @staticmethod
    def _get_attribute_data(value_type: ValueType, attribute_data_series: Series) -> Dict:
        """Generates metadata about a specific attribute based on a supplied dataframe

        :param value_type:                          DataType of the values in the series
        :param attribute_data_series:              Pandas Series of data for the specific attribute
        :return: Dictionary of attribute metadata and values inferred from the pandas series
        """
        if attribute_data_series.isnull().any():
            raise UserValueError("Column contains null value. Null values are not supported in reference or "
                                 f"inference data, please replace before proceeding.")

        cnt_distinct = len(attribute_data_series.unique())
        pct_distinct = cnt_distinct / float(len(attribute_data_series))
        attribute_data: Dict[str, Optional[Any]] = {'is_unique': True if pct_distinct == 1.0 else False}

        categorical = False
        if value_type == ValueType.Timestamp:
            categorical = False

        elif (value_type == ValueType.Float and
              cnt_distinct == 2 and
              attribute_data_series.max() == 1 and
              attribute_data_series.min() == 0):
            # exactly [0.0, 1.0]
            categorical = True

        elif value_type == ValueType.Float:
            categorical = False
        elif value_type == ValueType.Integer and cnt_distinct <= 20:
            categorical = True
        elif value_type == ValueType.Integer:
            categorical = False
        elif value_type == ValueType.Unstructured_Text:
            categorical = True
        else:
            categorical = True

        # even if unstructured text attributes aren't unique, we don't want categories
        if categorical and not attribute_data["is_unique"] and not value_type == ValueType.Unstructured_Text:
            attribute_data["categories"] = [AttributeCategory(value=str(cat)) for cat in
                                            list(set(attribute_data_series))]
        elif not categorical:
            attribute_data["min_range"] = core_util.NumpyEncoder.convert_value(attribute_data_series.min())
            attribute_data["max_range"] = core_util.NumpyEncoder.convert_value(attribute_data_series.max())

        attribute_data["categorical"] = categorical

        return attribute_data

    @staticmethod
    def _to_arthur_dtype( dtype: str) -> ValueType:
        """Select an :py:class:`.DataType` based on a pandas dtype

        :param dtype: the pandas dtype
        :return: The :py:class:`.DataType` corresponding to the pandas dtype
        """

        if dtype in ["string", "object"]:
            return ValueType.String
        elif dtype in ["bool", "boolean"]:
            return ValueType.Boolean
        elif re.search("u?int*", dtype, flags=re.I) or re.search("u?Int*", dtype, flags=re.I):
            return ValueType.Integer
        elif re.search("float*", dtype, flags=re.I):
            return ValueType.Float
        elif re.search("(datetime|timestamp).*", dtype, flags=re.I):
            return ValueType.Timestamp

    @arthur_excepted("failed to parse dataframe")
    def from_dataframe(self, data: Union[DataFrame, Series], stage: Stage) -> None:
        """ Attempt to autogenerate attributes based on input data. 
        
        Note that this does *not* automatically set reference
        data; this method only reads the passed-in data, and then infers attribute names, types, etc. and sets them up 
        within the ArthurModel. 

        For PredictedValue and GroundTruth stages, use the correct `add_<modeltype>_output_attributes()` method instead.

        :param data: the data to infer attribute metadata from
        :param stage: :py:class:`.Stage` of the data
        :return: None
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """

        if stage == Stage.PredictedValue or stage == Stage.GroundTruth:
            raise MethodNotApplicableError("Use either add_regression_output_attributes(), "
                                           "add_multiclass_classifier_output_attributes(), "
                                           "or add_binary_classifier_output_attributes() to add output attributes.")

        if isinstance(data, DataFrame):
            df = data
        elif isinstance(data, Series):
            df = data.to_frame()
        else:
            raise UserTypeError("Unsupported data type: a pandas.DataFrame or pandas.Series is required")

        for idx, column in enumerate(df.columns):

            series = core_util.standardize_pd_obj(df[column], dropna=True, replacedatetime=False)
            value_type = self._to_arthur_dtype(series.dtype.name)

            try:
                attribute_data = self._get_attribute_data(value_type, series)
            except ArthurUserError as e:
                raise ArthurUserError(f"Error in column: {column}: {str(e)}")

            if value_type == ValueType.String and self.input_type == InputType.NLP:
                value_type = ValueType.Unstructured_Text
                # even if unstructured text attributes are not unique we don't want to store categories
                attribute_data['categories'] = None

            attribute_data["position"] = idx
            arthur_attribute = ArthurAttribute(
                name=column,
                stage=stage,
                value_type=value_type,
                **attribute_data
            )
            self._add_attribute_to_model(arthur_attribute)

    def _add_attribute_to_model(self, attribute: ArthurAttribute) -> ArthurAttribute:
        """Adds an already-made ArthurAttribute to the model. 

        :param attribute: ArthurAttribute Object to add to the model
        :return: ArthurAttribute Object
        """
        if self.attributes is None:
            self.attributes = [attribute]
        else:
            self.attributes.append(attribute)

        return attribute

    @arthur_excepted("failed to add attribute")
    def add_attribute(self,
                      name: Optional[str] = None,
                      value_type: Optional[ValueType] = None,
                      stage: Optional[Stage] = None,
                      label: Optional[str] = None,
                      position: Optional[int] = None,
                      categorical: bool = False,
                      min_range: Optional[Union[float, int]] = None,
                      max_range: Optional[Union[float, int]] = None,
                      monitor_for_bias: bool = False,
                      categories: Optional[List[Union[str, AttributeCategory]]] = None,
                      bins: Optional[List[Union[NumberType, AttributeBin]]] = None,
                      is_unique: bool = False,
                      is_positive_predicted_attribute: bool = False,
                      attribute_link: Optional[str] = None,
                      arthur_attribute: Optional[ArthurAttribute] = None,
                      gt_pred_attrs_map: Optional[Dict] = None) -> List[ArthurAttribute]:
        """Adds a new attribute to the model and returns the attribute. 
        
        Also validates that Stage is not PredictedValue or GroundTruth. Additionally, attribute names must contain
        only letters, numbers, and underscores, and cannot begin with a number.

        :param attribute_link: Only applicable for `GroundTruth` or `PredictedValue` staged attributes.
        If stage is equal to `GroundTruth`, this represents the associated `PredictedValue` attribute and vice versa
        :param is_positive_predicted_attribute: Only applicable for `PredictedValue` attributes on a Binary
        Classification model. Should be set to `True` on the positive predicted value attribute.
        :param is_unique: Boolean value used to signal if the values of this attribute are unique.
        :param bins: List of bin cut-offs used to discretize continuous attributes. Use `None` as an open ended value.
                     `[None, 18, 65, None]` represents the three following bins:
                     `value < 18, 18 < value < 65, value > 65`
        :param monitor_for_bias: boolean value set to `True` if the attribute should be monitored for bias
        :param max_range: Max value for a continuous attribute
        :param min_range: Min value for a continuous attribute
        :param categorical: Boolean value set to `True` if the attribute has categorical values.
        :param position: The array position of attribute within the stage. Required in the PREDICT_FUNCTION_INPUT stage.
        :param label: Label for attribute. If attribute has an encoded name, a more readable label can be set.
        :param stage: :class:`arthurai.common.constants.Stage` of this attribute in the model pipeline
        :param value_type: :class:`arthurai.common.constants.ValueType` associated with this attributes values
        :param name: Name of the attribute. Attribute names can only contain alpha-numeric characters and underscores
        and cannot start with a number.
        :param categories: [Only for Categorical Attributes] If the attribute is categorical, this will contain the
        attribute's categories. It is required only if the categorical flag is set to true.
        :param arthur_attribute: Optional ArthurAttribute to add to the model
        :param gt_pred_attrs_map: .. deprecated:: version 2.0.0 Use `ArthurModel.add_[model_type]_output_attributes()` instead
        
        :return: ArthurAttribute Object
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if arthur_attribute is None:
            if stage is None or value_type is None or name is None:
                raise MissingParameterError("At minimum stage, value_type, and name must be provided for an attribute")
            elif stage == Stage.PredictedValue or stage == Stage.GroundTruth or gt_pred_attrs_map is not None:
                raise MethodNotApplicableError("Use either add_regression_output_attributes(), "
                                               "add_multiclass_classifier_output_attributes(), "
                                               "or add_binary_classifier_output_attributes() to add output attributes.")

            if not self._validate_attribute_name(name):
                raise UserValueError("Invalid attribute name: must contain only numbers, letters, and underscores, "
                                     "and cannot start with a number.")

            # Add categories or bins if supplied
            attribute_categories = [AttributeCategory(value=c) if not isinstance(c, AttributeCategory) else c
                                    for c in categories] if categories else None

            attribute_bins: Optional[List[Any]] = None
            if bins and not isinstance(bins[0], AttributeBin):
                attribute_bins = [AttributeBin(bins[i], bins[i + 1]) for i in range(len(bins) - 1)]
            elif bins and isinstance(bins[0], AttributeBin):
                attribute_bins = bins

            attribute_to_add = ArthurAttribute(
                name=name,
                value_type=value_type,
                stage=stage,
                label=label,
                position=position,
                categorical=categorical,
                min_range=min_range,
                max_range=max_range,
                monitor_for_bias=monitor_for_bias,
                categories=attribute_categories,
                bins=attribute_bins,
                is_unique=is_unique,
                is_positive_predicted_attribute=is_positive_predicted_attribute,
                attribute_link=attribute_link
            )
            return [self._add_attribute_to_model(attribute_to_add)]

        else:
            self._validate_attribute_name(arthur_attribute.name)
            return [self._add_attribute_to_model(arthur_attribute)]

    @staticmethod
    def _validate_attribute_name(attribute_name: str) -> bool:
        """Checks that attribute name is valid. 

        :param attribute_name: name of the attribute to add to the model
        :return: boolean indicating whether the name is valid. 
        """

        if bool(re.compile('\d').search(attribute_name[0])):  # \d is all the digits 0-9
            print("Attribute name cannot start with a number.")
            return False

        if bool(re.compile('\W').search(attribute_name)):  # \W is the complement of all alphanumeric characters and _.
            print("Attribute name can only contain numbers, letters, and underscores.")
            return False

        return True

    @arthur_excepted("failed to add regression output attributes")
    def add_regression_output_attributes(self, pred_to_ground_truth_map: Dict[str, str], value_type: ValueType) \
            -> Dict[str, ArthurAttribute]:
        """ Registers ground truth and predicted attribute parameters for regression models. 
        This function will register two ArthurAttribute objects to the model, a predicted value attribute 
        and ground truth attribute.

        :param pred_to_ground_truth_map: Map of predicted value attributes to their corresponding ground truth attribute names.
                                  The names provided in the dictionary will be used to register the one-hot encoded
                                  version of the attributes.
        :param value_type: Value type of regression model output attribute (usually either ValueType.Integer or ValueType.Float)
        
        :return: Mapping of added attributes string name -> ArthurAttribute Object
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        attributes_added = {}
        for pred_attr, gt_attr in pred_to_ground_truth_map.items():
            arthur_gt_attr = ArthurAttribute(
                name=gt_attr,
                stage=Stage.GroundTruth,
                value_type=value_type,
                categorical=False,
                attribute_link=pred_attr,
                position=0
            )
            arthur_pred_attr = ArthurAttribute(
                name=pred_attr,
                stage=Stage.PredictedValue,
                value_type=value_type,
                categorical=False,
                attribute_link=gt_attr,
                position=0
            )
            self._add_attribute_to_model(arthur_gt_attr)
            self._add_attribute_to_model(arthur_pred_attr)
            attributes_added[arthur_pred_attr.name] = arthur_pred_attr
            attributes_added[arthur_gt_attr.name] = arthur_gt_attr
        return attributes_added

    @arthur_excepted("failed to add output attributes")
    def add_multiclass_classifier_output_attributes(self, pred_to_ground_truth_map: Dict[str, str]) \
            -> Dict[str, ArthurAttribute]:
        """Registers ground truth and predicted attribute parameters. This function will create a predicted value and
        ground truth attribute for each mapping specified in pred_to_ground_truth_map.

        :param pred_to_ground_truth_map: Map of predicted value attributes to their corresponding ground truth attribute names.
                                  The names provided in the dictionary will be used to register the one-hot encoded
                                  version of the attributes. Ensure the ordering of items in this dictionary is an accurate
                                  representation of how model predictions (probability vectors) will be generated.
        :return: Mapping of added attributes string name -> ArthurAttribute Object
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """

        if len(pred_to_ground_truth_map) == 2:
            raise MethodNotApplicableError(
                "To add output attributes to a multiclass model with two output attributes, use `add_binary_classifier_output_attributes`.")

        attributes_added = {}
        for (index, (pred_attr, gt_attr)) in enumerate(pred_to_ground_truth_map.items()):
            arthur_gt_attr = ArthurAttribute(
                name=gt_attr,
                stage=Stage.GroundTruth,
                value_type=ValueType.Integer,
                categorical=True,
                categories=[AttributeCategory(value="0"), AttributeCategory(value="1")],
                attribute_link=pred_attr,
                position=index
            )
            arthur_pred_attr = ArthurAttribute(
                name=pred_attr,
                stage=Stage.PredictedValue,
                value_type=ValueType.Float,
                min_range=0,
                max_range=1,
                attribute_link=gt_attr,
                position=index
            )
            self._add_attribute_to_model(arthur_gt_attr)
            self._add_attribute_to_model(arthur_pred_attr)
            attributes_added[arthur_pred_attr.name] = arthur_pred_attr
            attributes_added[arthur_gt_attr.name] = arthur_gt_attr
        return attributes_added

    @arthur_excepted("failed to add output attributes")
    def add_binary_classifier_output_attributes(self, positive_predicted_attr: str,
                                                pred_to_ground_truth_map: Dict[str, str],
                                                threshold: float = 0.5) -> Dict[str, ArthurAttribute]:
        """Registers ground truth and predicted attribute parameters and their thresholds. 
        
        This function will create a predicted value and ground truth attribute for each mapping specified in 
        pred_to_ground_truth_map. 
        
        For binary models, `GroundTruth` is always an integer, and `PredictedAttribute` is always a float. 
        Additionally, `PredictedAttribute` is expected to be a probability (e.g. the output of a scikit-learn
        model's `predict_proba` method), rather than a classification to 0/1. 

        This assumes that separate columns for predicted values and ground truth values have already been created, 
        and that they have both been broken into two separate (pseudo-onehot) columns: for example, the column 
        `ground_truth_label` becomes `ground_truth_label=0` and `ground_truth_label=1`, and the column `pred_prob`
        becomes `pred_prob=0` and `pred_prob=1`. The pandas function `pd.get_dummies()` can be useful for reformatting
        the ground truth column, but be sure that the datatype is specified correctly as an int.

        :param positive_predicted_attr: string name of the predicted attribute to register as the positive predicted attribute
        :param pred_to_ground_truth_map: Map of predicted value attributes to their corresponding ground truth attribute names.
                                  The names provided in the dictionary will be used to register the one-hot encoded
                                  version of the attributes. For example: `{'pred_0': 'gt_0', 'pred_1': 'gt_1'}`,
                                  Ensure the ordering of items in this dictionary is an accurate
                                  representation of how model predictions (probability vectors) will be generated.
        :param threshold: Threshold to use for the classifier model, defaults to 0.5
        
        :return: Mapping of added attributes string name -> ArthurAttribute Object
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if len(pred_to_ground_truth_map) != 2:
            raise UserValueError((f"Binary classifiers must have two output attributes, but pred_to_ground_truth_map "
                                  f"has {len(pred_to_ground_truth_map)}. To add more than two output attributes to a "
                                  f"multiclass model, use add_multiclass_classifier_output_attributes()."))
        if positive_predicted_attr not in pred_to_ground_truth_map.keys():
            raise UserValueError((f"The positive_predicted_attr must be included in the mapping "
                                  f"pred_to_ground_truth_map. positive_predicted_attr {positive_predicted_attr} not "
                                  f"found in pred_to_ground_truth_map keys: {pred_to_ground_truth_map.keys()}."))
        attributes_added = {}
        for (index, (pred_attr, gt_attr)) in enumerate(pred_to_ground_truth_map.items()):
            is_pos_pred_attr = True if positive_predicted_attr == pred_attr else False
            arthur_gt_attr = ArthurAttribute(
                name=gt_attr,
                stage=Stage.GroundTruth,
                value_type=ValueType.Integer,
                categorical=True,
                categories=[AttributeCategory(value="0"), AttributeCategory(value="1")],
                attribute_link=pred_attr,
                position=index
            )
            arthur_pred_attr = ArthurAttribute(
                name=pred_attr,
                stage=Stage.PredictedValue,
                value_type=ValueType.Float,
                min_range=0,
                max_range=1,
                attribute_link=gt_attr,
                is_positive_predicted_attribute=is_pos_pred_attr,
                position=index
            )
            self._add_attribute_to_model(arthur_gt_attr)
            self._add_attribute_to_model(arthur_pred_attr)
            attributes_added[arthur_pred_attr.name] = arthur_pred_attr
            attributes_added[arthur_gt_attr.name] = arthur_gt_attr
        self.classifier_threshold = threshold
        return attributes_added

    @arthur_excepted("failed to get attribute")
    def get_attribute(self, name: str, stage: Optional[Stage] = None) -> ArthurAttribute:
        """Retrieves an attribute by name and stage

        :param name: string name of the attribute to retrieve
        :param stage: Optional `Stage` of attribute to retrieve
        
        :return: ArthurAttribute Object
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if self.attributes is None:
            raise MethodNotApplicableError("model does not have any attributes")
        for attr in self.attributes:
            if stage is not None and attr.name == name and attr.stage == Stage:
                return attr
            elif attr.name == name:
                return attr

        raise UserValueError(f"Attribute with name: {name} in stage: {stage} does not exist")

    @arthur_excepted("failed to get attributes")
    def get_attributes(self, stage: Optional[Stage]) -> Optional[List[ArthurAttribute]]:
        """Returns a list of attributes for the specified stage

        :param stage: :class:`arthurai.common.constants.Stage` to filter by
        :return: List of :class:`arthurai.attributes.ArthurAttribute`
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if self.attributes is None:
            return None
        return [attr for attr in self.attributes if stage and attr.stage == stage]

    @arthur_excepted("failed to get attribute names")
    def get_attribute_names(self, stage: Optional[Stage]) -> List[str]:
        """ Returns a list of all attribute names. If stage is supplied it will only return the attribute names which
        are in the specified stage.

        :param stage: :class:`arthurai.common.constants.Stage` to filter by
        :return: List of string attribute names
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if self.attributes is None:
            raise MethodNotApplicableError("model does not have any attributes")
        return [attr.name for attr in self.attributes if stage and attr.stage == stage]

    @arthur_excepted("failed to rename attribute")
    def rename_attribute(self, old_name: str, new_name: str, stage: Stage) -> ArthurAttribute:
        """Renames an attribute by name and stage

        :param old_name: string name of the attribute to rename
        :param new_name: string new name of the attribute
        :param stage: `Stage` of attribute
        
        :return: ArthurAttribute Object
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        attribute = self.get_attribute(old_name, stage)
        attribute.name = new_name

        return attribute

    @arthur_excepted("failed to set attribute labels")
    def set_attribute_labels(self, attribute_name: str, labels: Dict[Union[int, str], str],
                             attribute_stage: Optional[Stage] = None):
        """
        Sets labels for individual categories of a specific attribute
        :param attribute_name: Attribute name to set the categories labels
        :param attribute_stage: Optional stage of the attribute which is being updated
        :param labels: Dictionary where the key is the categorical value and the value is the string categorical label
        :return: None
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        attr_to_update = self.get_attribute(name=attribute_name, stage=attribute_stage)
        categories = []
        for cat_value, cat_label in labels.items():
            categories.append(AttributeCategory(value=cat_value, label=cat_label))
        attr_to_update.categories = categories

    @arthur_excepted("failed to generate summary")
    def review(self, stage: Stage = None, props: Optional[List[str]] = None, print_df=True) -> None:
        """Prints a summary of the properties of all attributes in the model.

        :param stage: restrict the output to a particular :py:class:`.Stage` (defaults to all stages)
        :param props: a list of properties to display in the summary
                   valid properties are data_type, categorical, is_unique, categories, cutoffs, range, monitor_for_bias, position
                   (defaults to data_type, categorical, is_unique)
        :param print_df: boolean value whether to print df or return it, defaults to False

        :return: Optional[pandas.Dataframe]
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """

        if props is None:
            props = ["value_type", "categorical", "is_unique", "categories", "bins", "range", "monitor_for_bias"]

        attributes: Optional[List[ArthurAttribute]] = None
        if stage is None:
            attributes = self.attributes
            display_items = ["name", "stage"] + props
        else:
            attributes = self.get_attributes(stage=stage)
            display_items = ["name"] + props
        if attributes is None:
            raise MissingParameterError("model does not have any attributes")

        result_df = DataFrame(columns=display_items)
        for attribute in attributes:
            row: Dict[str, Union[str, List[str]]] = {}
            for item in display_items:
                if item == 'range':
                    row[item] = f"[{attribute.min_range}, {attribute.max_range}]"
                elif item == 'categories':
                    string_categories = []
                    categories = [] if attribute.categories is None else attribute.categories
                    for cat in categories:
                        if cat.label is not None:
                            string_categories.append("{" + f"label: {cat.label}, value: {cat.value}" + "}")
                        else:
                            string_categories.append("{" + f"value: {cat.value}" + "}")
                    row[item] = string_categories
                else:
                    row[item] = attribute.__getattribute__(item)
            result_df = result_df.append(row, ignore_index=True)

        if print_df:
            try:
                display(result_df) # type: ignore
            except NameError:
                print(result_df)
        else:
            return result_df

    @arthur_excepted("failed to set predict function input order")
    def set_predict_function_input_order(self, attributes: List[str]) -> None:
        """Sets the expected order of attributes used by the prediction function.

        :param attributes: a list of attribute names

        :return: None
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        for idx, name in enumerate(attributes):
            attribute = self.get_attribute(name, Stage.ModelPipelineInput)
            attribute.position = idx

    @arthur_excepted("failed to set attribute as sensitive")
    def set_attribute_as_sensitive(self, attribute_name: str, attribute_stage: Optional[Stage] = None) -> None:
        """Sets the passed-in attribute to be sensitive by setting `attr.monitor_for_bias` = True.

        You will need to call `self.save()` or `self.update()` after this method; we do not
        automatically call the API in this method.

        :param attribute_name: Name of attribute to set as sensitive.
        :param attribute_stage: Stage of attribute to set as sensitive.

        :return: None

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        attribute = self.get_attribute(attribute_name, attribute_stage)

        attribute.monitor_for_bias = True

    @arthur_excepted("failed to enable explainability")
    def enable_explainability(self, df: DataFrame,
                              project_directory: str,
                              user_predict_function_import_path: str,
                              streaming_explainability_enabled: Optional[bool] = True,
                              requirements_file: str = "requirements.txt",
                              python_version: str = sys.version,
                              sdk_version: str = __version__,
                              model_server_num_cpu: Optional[str] = None,
                              model_server_memory: Optional[str] = None,
                              model_server_max_replicas: Optional[int] = None,
                              inference_consumer_num_cpu: Optional[str] = None,
                              inference_consumer_memory: Optional[str] = None,
                              inference_consumer_thread_pool_size: Optional[int] = None,
                              inference_consumer_score_percent: Optional[float] = None,
                              explanation_nsamples: Optional[int] = None,
                              explanation_algo: Optional[str] = None,
                              ignore_dirs: List = None) -> None:
        """Enable explainability for this model.

        :param df:                                a dataframe containing the :py:class:`.Stage.ModelPipelineInput` values for this model
        :param project_directory:                 the name of the directory containing the model source code
        :param user_predict_function_import_path: the name of the file that implements or wraps the predict function
        :param streaming_explainability_enabled:  Defaults to true. flag to turn on streaming explanations which will explain every inference sent to the platform. If false, explanations will need to be manually generated for each inference via the Arthur API. Set to false if worried about compute cost.
        :param requirements_file:                 the name of the file that contains the pip requirements (default: requirements.txt)
        :param python_version:                    the python version (default: sys.version)
        :param sdk_version:                       the version of the sdk to initialize the model servier with
        :param model_server_num_cpu:              string number of CPUs to provide to model server docker container. If not provided, 1 CPU is used. Specified in the format of Kubernetes CPU resources. '1', '1.5', '100m', etc. (default: None)
        :param model_server_memory:               The amount of memory to allocate to the model server docker container. Provided in the format of kubernetes memory resources "1Gi" or "500Mi" (default: None).
        :param model_server_max_replicas          The max number of model servers to create
        :param inference_consumer_num_cpu:        string number of CPUs to provide to inference consumer docker container. If not provided, 1 CPU is used. Specified in the format of Kubernetes CPU resources. '1', '1.5', '100m', etc. (default: '1')
        :param inference_consumer_memory:         The amount of memory to allocate to the model server docker container. Provided in the format of kubernetes memory resources "1Gi" or "500Mi" (default: '1G').
        :param inference_consumer_thread_pool_size  The number of inference consumer workers, this determines how many requests to the model server can be made in parallel. Default of 5. If increasing, CPU should be increased as well.
        :param inference_consumer_score_percent   What percent of inferences should get scored. Should be a value between 0.0 and 1.0. Default 1.0 (everything is scored)
        :param explanation_nsamples:              number of predictions to use in the explanation. For SHAP the default is 2048 + 2(num features). For LIME, the default is 5000. (default: None)
        :param explanation_algo:                  the algorithm to use for explaining inferences. Valid values are 'lime' and 'shap'. Defaults to 'lime'.
        :param ignore_dirs:                       a list of directories within the project_directory that you do not want to include when uploading the model.  Path is relative to project_directory.
        :return: None
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if self.input_type == InputType.NLP and self.text_delimiter is None:
            raise MissingParameterError("Must set a text delimiter for NLP models prior to enabling explainability")
        explainability_config = dict(
            df=df,
            project_directory=project_directory,
            ignore_dirs=ignore_dirs if ignore_dirs else [],
            user_predict_function_import_path=user_predict_function_import_path,
            streaming_explainability_enabled=streaming_explainability_enabled,
            requirements_file=requirements_file,
            python_version=python_version,
            sdk_version=sdk_version,
            explanation_nsamples=explanation_nsamples,
            explanation_algo=explanation_algo,
            model_server_num_cpu=model_server_num_cpu,
            model_server_memory=model_server_memory,
            model_server_max_replicas=model_server_max_replicas,
            inference_consumer_num_cpu=inference_consumer_num_cpu,
            inference_consumer_memory=inference_consumer_memory,
            inference_consumer_thread_pool_size=inference_consumer_thread_pool_size,
            inference_consumer_score_percent=inference_consumer_score_percent
        )
        return self.update_enrichment(Enrichment.Explainability, True, explainability_config)

    def _check_model_save(self, msg="You must save the model before sending"):
        if self.id is None:
            raise MethodNotApplicableError(msg)

    @arthur_excepted("failed to send inferences")
    def send_inferences(self, inferences: List[Dict[str, Any]], fail_silently: bool = False):
        """Uploads an list of json inferences of the format below.
        
        All inferences should follow the column format specified in `add_<modeltype>_classifier_output_attributes()`.
        Additionally, all inferences must have `partner_inference_id` and `inference_timestamp` fields. If
        `ground_truth_data` is supplied then `ground_truth_timestamp` is required. All timestamps should be in UTC
        ISO 8601 format. `inference_data` Should contain a key/value pair for every registered attribute name and value.

        :param inferences: a list of json inference objects to upload:

         .. code-block:: JSON

            [
                {
                    "inference_timestamp": "string",
                    "partner_inference_id": "string",
                    "batch_id": "string",
                    "inference_data": {},
                    "ground_truth_timestamp": "string",
                    "ground_truth_data": {}
                }
            ]
        :param fail_silently: if True, log failed inferences but do not raise an exception

        :return: Upload status response in the following format:

         .. code-block:: JSON

            {
              "counts": {
                "success": 0,
                "failure": 0,
                "total": 0
              },
              "results": [
                {
                  "message": "success",
                  "status": 200
                }
              ]
            }

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        self._check_model_save(msg="Must save model before sending inferences.")

        inferences = arthur_util.format_timestamps(inferences)
        
        for i in range(len(inferences)):
            inferences[i]["inference_data"] = self._replace_nans_and_infinities_in_dict(inferences[i]["inference_data"])
            inferences[i]['inference_data'] = self._convert_numpy_to_native(inferences[i]["inference_data"])
            if 'ground_truth_data' in inferences[i]:
                inferences[i]['ground_truth_data'] = self._convert_numpy_to_native(inferences[i]["ground_truth_data"])

        url = f"{self._client.base_path}/models/{self.id}/inferences"
        resp = self._client.post(url, data=inferences, return_raw_response=True)

        user_failures, internal_failures = \
            validation.validate_multistatus_response_and_get_failures(resp, raise_on_failures=(not fail_silently))
        if fail_silently and (len(user_failures) > 0) or (len(internal_failures) > 0):
            message = "not all inferences succeeded!"
            message += f" user failures: {user_failures}." if len(user_failures) > 0 else ""
            message += f" internal failures: {internal_failures}." if len(internal_failures) > 0 else ""
            logger.error(message)

        return resp.json()

    def _format_inference_request(self, inference_timestamp: Union[str, datetime],
                                  partner_inference_id: Optional[str] = None,
                                  model_pipeline_input=None,
                                  non_input_data=None,
                                  predicted_value=None,
                                  ground_truth=None):
        """takes in an inference to send following the old sdk contract and converts the data to the request body
        of the new api format.


        :param inference_timestamp: a mapping of the name of ground truth attributes to their value
        :param partner_inference_id: an external id (partner_inference_id) to assign to the inferences
        :param model_pipeline_input: a mapping of the name of pipeline input attributes to their value
        :param non_input_data: a mapping of the name of non-input data attributes to their value
        :param predicted_value: a mapping of the name of predicted value attributes to their value
        :param ground_truth: a mapping of the name of ground truth attributes to their value

        :return: dictionary object which can be used to send the inference
        """
        if model_pipeline_input is None:
            model_pipeline_input = {}
        if non_input_data is None:
            non_input_data = {}
        if predicted_value is None:
            predicted_value = {}

        model_pipeline_input.update(predicted_value)
        model_pipeline_input.update(non_input_data)
        inference = {
            "inference_timestamp": inference_timestamp,
            "partner_inference_id": partner_inference_id,
            "inference_data": model_pipeline_input,
        }

        if ground_truth is not None:
            inference["ground_truth_timestamp"] = inference_timestamp
            inference["ground_truth_data"] = ground_truth

        return inference

    @staticmethod
    def _replace_nans_and_infinities_in_dict(dict_to_update) -> Optional[Dict]:
        if dict_to_update is None:
            return None

        return dict(
            (k, v if ((type(v) != float or not np.isnan(v)) and v != np.Inf and v != -np.inf and v != np.inf) else None)
            for k, v in dict_to_update.items())

    @staticmethod
    def _convert_numpy_to_native(dict_to_update) -> Dict:
        final_dict = {}
        for k, v in dict_to_update.items():
            if isinstance(v, np.generic):
                final_dict[k] = v.item()
            else:
                final_dict[k] = v
        return final_dict

    @arthur_excepted("failed to send inference")
    def send_inference(self, inference_timestamp: Union[str, datetime],
                       partner_inference_id: str = "",
                       model_pipeline_input=None,
                       non_input_data=None,
                       predicted_value=None,
                       ground_truth=None):
        """Uploads an inference with or without ground truth; this sends a single inference.

        All inferences should follow the column format specified in `add_<modeltype>_classifier_output_attributes()`. 
        Additionally, external_id and inference_timestamp must be provided. 

        :param inference_timestamp: timestamp for inference to send; generated by external partner (not Arthur)
        :param partner_inference_id: an external id (partner_inference_id) to assign to the inferences
        :param model_pipeline_input: a mapping of the name of pipeline input attributes to their value
        :param non_input_data: a mapping of the name of non-input data attributes to their value
        :param predicted_value: a mapping of the name of predicted value attributes to their value
        :param ground_truth: a mapping of the name of ground truth attributes to their value
        
        :return: Upload status response in the following format:

         .. code-block:: JSON

            {
                "counts": {
                    "success": 0,
                    "failure": 0,
                    "total": 0
                },
                "results": [
                    {
                        "message": "success",
                        "status": 200
                    }
                ]
            }

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if predicted_value is None:
            predicted_value = {}
        if non_input_data is None:
            non_input_data = {}
        if model_pipeline_input is None:
            model_pipeline_input = {}

        self._check_model_save(msg="Must save model before sending inferences.")

        inference = self._format_inference_request(
            inference_timestamp=inference_timestamp,
            partner_inference_id=partner_inference_id,
            model_pipeline_input=ArthurModel._replace_nans_and_infinities_in_dict(model_pipeline_input),
            non_input_data=ArthurModel._replace_nans_and_infinities_in_dict(non_input_data),
            predicted_value=ArthurModel._replace_nans_and_infinities_in_dict(predicted_value),
            ground_truth=ArthurModel._replace_nans_and_infinities_in_dict(ground_truth)
        )
        return self.send_inferences([inference])

    @arthur_excepted("failed to update inference ground truths")
    def update_inference_ground_truths(self, ground_truths: List[Dict[str, Any]], fail_silently: bool = False):
        """Updates a list of inferences with the supplied ground truth values.
        Ground truth json must include fields "ground_truth_timestamp", "partner_inference_id",
        and "ground_truth_data".
        Field "ground_truth_data" is a dictionary with keys corresponding to attributes names
        from Stage.GroundTruth.

        :param ground_truths: a mapping of the name of a ground truth attribute name to its value

            ..  code-block:: JSON

                [
                  {
                    "ground_truth_timestamp": "string",
                    "partner_inference_id": "string",
                    "ground_truth_data": {
                      "gt_fraud": 1,
                      "gt_not_fraud": 0
                    }
                  }
                ]
        :param fail_silently: if True, log failed inferences but do not raise an exception

        :return: Upload status response in the following format:

            .. code-block:: JSON

                {
                    "counts": {
                        "success": 0,
                        "failure": 0,
                        "total": 0
                    },
                    "results": [
                        {
                            "message": "success",
                            "status": 200
                        }
                    ]
                }

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        self._check_model_save()

        ground_truths = arthur_util.format_timestamps(ground_truths)

        url = f"{self._client.base_path}/models/{self.id}/inferences"
        resp = self._client.patch(url, data=ground_truths, return_raw_response=True)
        user_failures, internal_failures = \
            validation.validate_multistatus_response_and_get_failures(resp, raise_on_failures=(not fail_silently))
        if fail_silently and (len(user_failures) > 0) or (len(internal_failures) > 0):
            message = "not all ground truth updates succeeded!"
            message += f" user failures: {user_failures}." if len(user_failures) > 0 else ""
            message += f" internal failures: {internal_failures}." if len(internal_failures) > 0 else ""
            logger.error(message)

        return resp.json()

    @arthur_excepted("failed to binarize")
    def binarize(self, attribute_value):
        """Creates a binary class probability based on classes defined in a :py:attr:`.ModelType.Multiclass` model.

        :param attribute_value: a mapping of the name of a predicted value attribute to its value

        :return: A two-value dictionary with probabilities for both predicted classes.
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """

        if len(attribute_value) > 1:
            raise UserValueError("A dictionary containing a key, value pair for one ground truth attribute is required")

        name, value = attribute_value.popitem()
        name = str(name)
        predicted_value = self.get_attributes(Stage.PredictedValue)
        if name not in [attr.name for attr in predicted_value]:
            raise UserValueError(f"Attribute {name} not found in {Stage.PredictedValue}")

        valid_type = lambda attr: attr.value_type == ValueType.Float and attr.min_range == 0.0 and attr.max_range == 1.0
        if len(predicted_value) == 2 and all([valid_type(attr) for attr in predicted_value]):
            return dict([(attr.name, value if attr.name == name else 1 - value) for attr in predicted_value])
        else:
            raise MethodNotApplicableError("This model is not a binary classification model.")

    @arthur_excepted("failed to one hot encode")
    def one_hot_encode(self, value):
        """Creates a one hot encoding of a class label based on classes defined in a :py:attr:`.ModelType.Multiclass` model.

        :param value: the ground truth value

        :return: A dictionary with a one hot encoding of possible ground truth values.
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        name = str(value)
        ground_truth = self.get_attribute_names(stage=Stage.GroundTruth)
        if name not in ground_truth:
            raise UserValueError(f"Attribute {name} not found in {Stage.GroundTruth}")
        elif self.output_type != OutputType.Multiclass:
            raise MethodNotApplicableError("This model is not a Multiclass model")
        else:
            return dict([(attr_name, 1 if attr_name == name else 0) for attr_name in ground_truth])

    def _send_parquet_files(self,
                            url: str,
                            file_name: str,
                            directory_path: Optional[str] = None,
                            data: Optional[Union[DataFrame, Series]] = None,
                            initial_form_data: Optional[Dict[str, Any]] = None,
                            retries: int = 0):

        self._check_model_save(msg="Must save model before uploading reference data")

        if directory_path is None and data is None:
            raise MissingParameterError("Either directory_path or data must be provided")

        if data is not None:
            if isinstance(data, DataFrame):
                filename = DatasetService.convert_dataframe(model_id=self.id, stage=None, df=data)
            elif isinstance(data, Series):
                filename = DatasetService.convert_dataframe(model_id=self.id, stage=None, df=data.to_frame())
            else:
                raise UserTypeError("Unsupported data type: a pandas.DataFrame or pandas.Series is required")
            directory_path = filename[:filename.rfind('/')]

        _, file_info = DatasetService.send_parquet_files_from_dir_iteratively(self,
                                                                              directory_path,
                                                                              url,
                                                                              file_name,
                                                                              additional_form_params=initial_form_data,
                                                                              retries=retries)
        return file_info

    @arthur_excepted("failed to set reference data")
    def set_reference_data(self, directory_path: Optional[str] = None, data: Optional[Union[DataFrame, Series]] = None):
        """Validates and sets the reference data for the given stage to the provided data.
        
        Either directory_path or data must be provided. Additionally, there must be 
        one column per `ModelPipelineInput` and `NonInput` attribute. 

        :param directory_path: file path to a directory of parquet files to upload for batch data
        :param data:  a DataFrame or Series containing the ground truth data

        :return: Returns a tuple, the first variable is the response from sending the reference set and the second
        is the response from closing the dataset.

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if directory_path is None and data is None:
            raise MissingParameterError("Either directory_path or data must be provided")

        if data is not None:
            data = core_util.standardize_pd_obj(data, dropna=False, replacedatetime=False)

        url = f"{self._client.base_path}/models/{self.id}/reference_data"
        res = self._send_parquet_files(url, 'reference_data.parquet', directory_path, data)

        ref_set_close_res = None
        if res[DatasetService.COUNTS][DatasetService.FAILURE] != 0:
            logger.warning(f"{res[DatasetService.COUNTS][DatasetService.FAILURE]} inferences failed to upload")
            logger.warning("Reference dataset auto-close was aborted because not all "
                           "inferences in the reference set were successfully uploaded")
        else:
            if data is None:
                files = core_util.retrieve_parquet_files(directory_path)
                num_rows = self._count_parquet_num_rows(files, check_cols=False)
            else:
                num_rows = len(data)

            if num_rows == 0:
                raise UserValueError("data provided does not have any rows")

            ref_set_close_res = self._close_datatset(f"{self._client.base_path}/models/{self.id}/reference_data", num_rows)
        return res, ref_set_close_res

    @arthur_excepted("failed to send batch ground truths")
    def send_batch_ground_truths(self, directory_path: Optional[str] = None,
                                 data: Optional[Union[DataFrame, Series]] = None):
        """ Uploads parquet files containing a columns for partner_inference_id, ground_truth_timestamp, and column(s) for
        inference ground truth

        :param directory_path: file path to a directory of parquet files containing ground truth data
        :param data:  a DataFrame or Series containing the reference data for the :py:class:`Stage`

        :return: Upload status response in the following format:

            .. code-block:: JSON

                {
                    "counts": {
                        "success": 0,
                        "failure": 0,
                        "total": 0
                    },
                    "results": [
                        {
                            "message": "success",
                            "status": 200
                        }
                    ]
                }
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """

        self._check_model_save()

        if directory_path is None and data is None:
            raise MissingParameterError("Either directory_path or data must be provided")

        url = f"{self._client.base_path}/models/{self.id}/inferences/file"
        return self._send_parquet_files(url, 'ground_truths.parquet', directory_path, data,
                                        retries=INFERENCE_DATA_RETRIES)

    @arthur_excepted("failed to send batch inferences")
    def send_batch_inferences(self,
                              batch_id: str,
                              directory_path: Optional[str] = None,
                              data: Optional[DataFrame] = None,
                              complete_batch: bool = True):
        """Validates and uploads parquet files containing columns for inference data, partner_inference_id,
        inference_timestamp, and optionally a batch_id and ground truth data. Either directory_path or data must be
        specified.

        The columns for predicted attributes should follow the column format specified in 
        `add_<modeltype>_classifier_output_attributes()`.  Additionally, `partner_inference_id`, 
        `inference_timestamp`, must be specified for all inferences.

        :param directory_path: file path to a directory of parquet files containing ground truth data
        :param data:  a DataFrame containing the reference data. 
        :param batch_id: string id for the batch to upload; if supplied, this will override any batch_id column
                      specified in the provided dataset
        :param complete_batch: Defaults to true and will automatically close a batch once it is sent

        :return: A tuple of the batch upload response and the close batch response.
        The batch upload response is in the following format:

            .. code-block:: JSON

                {
                    "counts": {
                        "success": 0,
                        "failure": 0,
                        "total": 0
                    },
                    "results": [
                        {
                            "message": "success",
                            "status": 200
                        }
                    ]
                }
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        initial_form_data = {'batch_id': (None, batch_id)}
        url = f"{self._client.base_path}/models/{self.id}/inferences/file"

        if directory_path is None and data is None:
            raise MissingParameterError("Either directory_path or data must be provided.")

        if data is not None:
            if isinstance(data, DataFrame):
                num_rows = len(data)
                self._check_batch_inference_columns(data.columns)
                data = core_util.standardize_pd_obj(data, dropna=False, replacedatetime=False)
            else:
                raise UserTypeError("Unsupported data type: a pandas.DataFrame is required")
        else:
            files = core_util.retrieve_parquet_files(directory_path)
            num_rows = self._count_parquet_num_rows(files, check_cols=True)

        if num_rows == 0:
            raise UserValueError("data provided does not have any rows")

        res = self._send_parquet_files(url, 'inferences.parquet', directory_path, data,
                                       initial_form_data=initial_form_data,
                                       retries=INFERENCE_DATA_RETRIES)
        batch_res = None
        if res[DatasetService.COUNTS][DatasetService.FAILURE] != 0:
            logger.warning(f"{res[DatasetService.COUNTS][DatasetService.FAILURE]} inferences failed to upload")
            if complete_batch:
                logger.warning("Batch auto-close was aborted because not all inferences were successfully uploaded")
        elif complete_batch:
            url = f"{self._client.base_path}/models/{self.id}/batches/{batch_id}"
            batch_res = self._close_datatset(url, num_rows)
        return res, batch_res

    def _count_parquet_num_rows(self, file_paths: List[str], check_cols: bool = False) -> int:
        num_rows = 0
        for file in file_paths:
            pqfile = pq.ParquetFile(file)
            num_rows += pqfile.metadata.num_rows
            cols = [pqfile.schema[i].name for i in range(len(pqfile.schema))]
            if check_cols:
                self._check_batch_inference_columns(cols)

        return num_rows

    def _close_datatset(self, url: str, num_inferences: Optional[int] = None) -> Dict:
        body: Dict[str, Union[str, int]] = {
            "status": "uploaded"
        }
        if num_inferences is not None:
            body["total_record_count"] = num_inferences

        batch_res = self._client.patch(url=url, data=body, return_raw_response=True, retries=INFERENCE_DATA_RETRIES)
        if batch_res.status_code != 200:
            batch_message = batch_res.json()
            logger.warning(f"dataset auto-close failed: {batch_message}")
        else:
            batch_message = "success"
        return {"dataset_close_result": batch_message}

    @arthur_excepted("failed to close batch")
    def close_batch(self, batch_id: str, num_inferences: Optional[int] = None) -> Dict:
        """Closes the specified batch, optionally can supply the number of inferences that are contained in the batch

        :param batch_id: String batch_id associated with the batch that will be closed
        :param num_inferences: Optional number of inferences that are contained in the batch
        :return: Response of the batch close rest call
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        url = f"{self._client.base_path}/models/{self.id}/batches/{batch_id}"
        return self._close_datatset(url, num_inferences)

    @staticmethod
    def _check_batch_inference_columns(columns):
        error_message = ""
        if 'inference_timestamp' not in columns:
            error_message += '`inference_timestamp` must be a column in the data. '
        if 'partner_inference_id' not in columns:
            error_message += '`partner_inference_id` must be a column in the data. '

        if error_message != "":
            raise MissingParameterError(error_message)

    @arthur_excepted("failed to delete explainer")
    def delete_explainer(self) -> None:
        """Spin down the model explainability server.

        :return: the server response

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if not hasattr(self, "_explainer"):
            raise MethodNotApplicableError(f"There is no explainability server to delete for model {self.id}")

        url = f"{self._client.base_path}/models/{self.id}/explainability"
        resp = self._client.delete(url, return_raw_response=True)

        validation.validate_response_status(resp, expected_status_code=HTTPStatus.NO_CONTENT)
        logger.info(f"Successfully removed explainability server for model {self.id}")

    @arthur_excepted("failed to archive model")
    def archive(self):
        """Archives the model with a DELETE request.

        :return: the server response
        :raise: Exception: the model has no ID, or the model has not been archived
        """
        self._check_model_save(msg="Cannot archive an unregistered model.")

        url = f"{self._client.base_path}/models/{self.id}"
        resp = self._client.delete(url, return_raw_response=True)
        validation.validate_response_status(resp, expected_status_code=HTTPStatus.NO_CONTENT)

    @arthur_excepted("failed to execute query")
    def query(self, body: Dict[str, Any], query_type='base'):
        """ Execute query against the model's inferences.
        For full description of possible functions, aggregations, and transformations, see
        https://docs.arthur.ai/api-documentation/query_guide.html
        For queries pertaining to datadrift metrics, please see
        https://docs.arthur.ai/api-documentation/query_endpoint_documentation.html#data-drift

        :param body: dict
        :param query_type: str Can be either 'base' or 'drift'

        .. code-block:: python

            body = {
                       "select":[
                          {"property":"batch_id"},
                          {
                             "function":"count",
                             "alias":"inference_count"
                          }
                       ],
                       "group_by":[
                          {"property":"batch_id"}
                       ]
                    }

        .. code-block:: python

            body = {
                       "select":[,
                          {"property":"batch_id"},
                          {
                             "function":"rate",
                             "alias":"positive_rate",
                             "parameters":{
                                "property":"predicted_1",
                                "comparator":"gt",
                                "value":0.75
                             }
                          }
                       ],
                       "group_by":[
                          {"property":"batch_id"}
                       ]
                    }


        :return: the query response as documented in https://docs.arthur.ai/api-documentation/query_guide.html
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        self._check_model_save(msg="You must save model before querying.")

        url = f"{self._client.base_path}/models/{self.id}/inferences/query"
        if query_type == "drift":
            url = f"{url}/data_drift"
        resp = self._client.post(url, data=body, return_raw_response=True)

        validation.validate_response_status(resp, expected_status_code=HTTPStatus.OK)

        return resp.json()["query_result"]

    @arthur_excepted("failed to create alert rule")
    def create_alert_rule(self,
                          bound: Optional[AlertRuleBound] = None,
                          threshold: Optional[NumberType] = None,
                          metric: Optional[Union[AlertMetric, str]] = None,
                          severity: Optional[AlertRuleSeverity] = None,
                          query: Optional[dict] = None,
                          lookback_period: Optional[NumberType] = None,
                          subsequent_alert_wait_time: Optional[NumberType] = None,
                          check_batch_idle_time: Optional[NumberType] = None,
                          attribute_id: Optional[str] = None,
                          attribute_name: Optional[str] = None,
                          enabled: bool = True,
                          alert_rule: Optional[AlertRule] = None) -> AlertRule:
        """
        :param bound:
        :param threshold:
        :param metric:
        :param severity:
        :param query:
        :param lookback_period:
        :param subsequent_alert_wait_time:
        :param check_batch_idle_time:
        :param attribute_id:
        :param attribute_name:
        :param enabled:
        :param alert_rule:
        :return: the created alert rule
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if alert_rule is not None:
            if not isinstance(alert_rule, AlertRule):
                raise UserTypeError("alert_rule supplied must be an AlertRule Object")
            # cannot have a set id for an alert rule that hasn't been created yet
            alert_rule.id = None
        else:
            alert_rule = AlertRule(
                bound=bound,
                threshold=threshold,
                metric=metric,
                severity=severity,
                query=query,
                lookback_period=lookback_period,
                subsequent_alert_wait_time=subsequent_alert_wait_time,
                check_batch_idle_time=check_batch_idle_time,
                attribute_id=attribute_id,
                attribute_name=attribute_name,
                enabled=enabled
            )

        url = f"{self._client.base_path}/models/{self.id}/alert_rules"
        resp = self._client.post(url, data=alert_rule.to_dict(), return_raw_response=True)

        validation.validate_response_status(resp, expected_status_code=HTTPStatus.OK)

        return AlertRule.from_dict(resp.json())

    @arthur_excepted("failed to get model alert rules")
    def get_model_alert_rules(self, page: int = 1, page_size: int = 20) -> List[AlertRule]:
        """Returns a paginated list of alert rules registered to this model

        :param page: page of alert rules to retrieve, defaults to 1
        :param page_size: number of alert rules to return per page, defaults to 20
        :return: List of :class:`arthurai.client.apiv3.AlertRule` objects
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        url = f"{self._client.base_path}/models/{self.id}/alert_rules?page_size={page_size}&page={page}"
        resp = self._client.get(url, return_raw_response=True)

        validation.validate_response_status(resp, expected_status_code=HTTPStatus.OK)
        if 'data' not in resp.json():
            raise ExpectedParameterNotFoundError("An error occurred when retrieving alert rules: {0}".format(resp.json()))

        alert_rules = []
        for rule in resp.json()['data']:
            alert_rules.append(AlertRule.from_dict(rule))
        return alert_rules

    @arthur_excepted("failed to update alert rule")
    def update_alert_rule(self, alert_rule: AlertRule, alert_rule_id: Optional[str] = None):
        """Updates alert rule fields included in the `alert_rule` object for the specified alert rule id. If the
        alert rules id field is present in the `alert_rule` parameter that is used otherwise `alert_rule_id`
        must be supplied

        :param alert_rule: Object which contains fields to update on the specified alert rule
        :param alert_rule_id: If the alert rule id is not specified in the `alert_rule_to_update` object then this
                              must be provided to determine which alert rule to update.
        :return: Updates alert rule object
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if alert_rule.id is None and alert_rule_id is None:
            raise MethodNotApplicableError("alert_rule_to_update must have a valid id, if the alert rule has not been "
                                           "created yet call model.create_alert_rule(...)")

        alert_rule_id = alert_rule.id if alert_rule.id is not None else alert_rule_id
        alert_rule.id = None

        url = f"{self._client.base_path}/models/{self.id}/alert_rules/{alert_rule_id}"
        resp = self._client.patch(url, data=alert_rule.to_dict(), return_raw_response=True)

        validation.validate_response_status(resp, expected_status_code=HTTPStatus.OK)

        return AlertRule.from_dict(resp.json())

    @arthur_excepted("failed to get enrichments")
    def get_enrichments(self) -> Dict[str, Any]:
        """Returns configuration for all enrichments.

        :return: Upload status response in the following format:

            .. code-block:: JSON

                {
                    "anomaly_detection": {
                        "enabled": false
                    },
                    "explainability": {
                        "config": {
                            "python_version": "3.7",
                            "sdk_version": "3.0.11",
                            "streaming_explainability_enabled": false,
                            "user_predict_function_import_path": "entrypoint",
                            "shap_expected_values": "[0.7674405187893311, 0.23255948121066888]",
                            "model_server_cpu": "2",
                            "model_server_memory": "1Gi",
                            "model_server_max_replicas": "5",
                            "explanation_nsamples": 1000,
                            "explanation_algo": "lime",
                            "inference_consumer_cpu": "100m",
                            "inference_consumer_memory": "512Mi",
                            "inference_consumer_score_percent": "1.0",
                            "inference_consumer_thread_pool_size": "1",
                            "service_account_id": "8231affb-c107-478e-a1b4-e24e7f1f6619"
                        },
                        "enabled": true
                    }
                }
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        url = f"{self._client.base_path}/models/{self.id}/enrichments"
        resp = self._client.get(url, return_raw_response=True)
        validation.validate_response_status(resp, expected_status_code=HTTPStatus.OK)
        return resp.json()

    @arthur_excepted("failed to get enrichment")
    def get_enrichment(self, enrichment: Enrichment) -> Dict[str, Any]:
        """Returns configuration for the specified enrichment.

        :param enrichment: Enrichment constant
        :return: Enrichment config

            .. code-block:: JSON

                {
                    "enabled": true,
                    "config": {
                        "python_version": "3.7",
                        "sdk_version": "3.0.11",
                        "streaming_explainability_enabled": false,
                        "user_predict_function_import_path": "entrypoint",
                        "shap_expected_values": "[0.7674405187893311, 0.23255948121066888]",
                        "model_server_cpu": "2",
                        "model_server_memory": "1Gi",
                        "model_server_max_replicas": "5",
                        "explanation_nsamples": 1000,
                        "explanation_algo": "lime",
                        "inference_consumer_cpu": "100m",
                        "inference_consumer_memory": "512Mi",
                        "inference_consumer_score_percent": "1.0",
                        "inference_consumer_thread_pool_size": "1",
                        "service_account_id": "8231affb-c107-478e-a1b4-e24e7f1f6619"
                    }
                }
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        url = f"{self._client.base_path}/models/{self.id}/enrichments/{enrichment}"
        resp = self._client.get(url, return_raw_response=True)
        validation.validate_response_status(resp, expected_status_code=HTTPStatus.OK)
        return resp.json()

    @arthur_excepted("failed to update enrichments")
    def update_enrichments(self, enrichment_configs: Dict[Union[str, Enrichment], Any]) -> Dict[str, Any]:
        """Update the configuration for 1 or more enrichments.
        See the enrichments guide at http://docs.arthur.ai/guides/enrichments.html.

        :param enrichment_configs: Dict containing the configuration for each enrichment

            .. code-block:: JSON

            {
                    "anomaly_detection": {
                        "enabled": false
                    },
                    "explainability": {
                        "config": {
                            "streaming_explainability_enabled": false,
                            "explanation_nsamples": 1000,
                            "explanation_algo": "lime",
                            "inference_consumer_score_percent": "1.0"
                        },
                        "enabled": true
                    }
                }
        :return: the resulting enrichments configuration
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """

        files = []
        # handle specifics for different enrichments
        if Enrichment.Explainability in enrichment_configs and enrichment_configs[Enrichment.Explainability]['enabled']:
            explanationPackager = ExplanationPackager(self, **enrichment_configs[Enrichment.Explainability]['config'])

            # check to see if user is trying to update files
            onePresent, allPresent = explanationPackager.contains_file_fields()
            if onePresent and not allPresent:
                raise MissingParameterError(f"If updating model files for explainability, all of the following fields must be present in the config {ExplanationPackager.FILE_FIELDS}")
            # package explainer files
            elif allPresent:
                explanationPackager.make_zip()
                explanationPackager.create()
                files += explanationPackager.get_request_files()
            # replace passed in config (has file specific fields) with actual config
            enrichment_configs[Enrichment.Explainability]["config"] = explanationPackager.get_request_config()

        # build and make request
        headers = {"Content-Type": "multipart/form-data"}
        url = f"{self._client.base_path}/models/{self.id}/enrichments"
        data = {"config": json.dumps(enrichment_configs)}
        # we need to make sure when we send no files, the config dict is still marked as multipart with a boundary
        # to do this, we trick requests into thinking there are files, but there aren't
        if files == []:
            files.append(('',''))

        resp = self._client.patch(url, data, files=files, headers=headers, return_raw_response=True)
        validation.validate_response_status(resp, expected_status_code=HTTPStatus.ACCEPTED)
        return resp.json()

    @arthur_excepted("failed to update enrichment")
    def update_enrichment(self, enrichment: Enrichment, enabled: bool, config: Dict[str, Any]) -> Dict[str, Any]:
        """Update the configuration for a single enrichment.
        :param enrichment: the enrichment to update
        :param enabled: whether the enrichment should be enabled or disabled
        :param config: the configuration for the enrichment
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        return self.update_enrichments({
            enrichment: {"enabled": enabled, "config": config}
        })



class ListableStrEnum:
    """
    This class operates similar to Enum but passes mypy type checking.
    The members are accessed directly rather than by `.name` or `.value`.
    """

    @classmethod
    def list(cls):
        """
        Lists all attributes in alphabetical order
        """
        members = [getattr(cls, attr) for attr in dir(cls)
                   if not callable(getattr(cls, attr)) and not attr.startswith("__")]
        return sorted(members)


class InputType(ListableStrEnum):
    Tabular = "TABULAR"
    Image = "IMAGE"
    NLP = "NLP"


class OutputType(ListableStrEnum):
    Regression = "REGRESSION"
    Multiclass = "MULTICLASS"
    Multilabel = "MULTILABEL"


class ValueType(ListableStrEnum):
    String = "STRING"
    Integer = "INTEGER"
    Float = "FLOAT"
    Image = "IMAGE"
    Boolean = "BOOLEAN"
    Timestamp = "TIMESTAMP"
    Unstructured_Text = "UNSTRUCTURED_TEXT"


class Stage(ListableStrEnum):
    ModelPipelineInput = "PIPELINE_INPUT"
    PredictFunctionInput = "PREDICT_FUNCTION_INPUT"
    PredictedValue = "PREDICTED_VALUE"
    NonInputData = "NON_INPUT_DATA"
    GroundTruth = "GROUND_TRUTH"


class TextDelimiter(ListableStrEnum):
    """ This class contains patterns that can be used
    as text_delimiter for NLP models.

    Definitions:

    NOT_WORD: `'\W+'` Splits on any character that is not a word. Ex:
              "this  is,aaaa,,,,,test!" = ["this", "is", "aaaa", "test", ""]

    WHITESPACE: `'\s+` Splits on whitespace. Ex:
              "this  is,a test! " = ["this", "is,a", "test!", ""]

    COMMA: `","` Splits on a single comma. Ex:
              "this ,is,,a,test" = ["this ", "is", "", "a", "test"]

    COMMA_PLUS: `",+"` Splits on one or more commas. Ex:
              "this ,is,,a,test" = ["this ", "is", "a", "test"]

    PIPE: `"\|"` Splits on a single pipe. Ex:
              "this |is||a|test" = ["this ", "is", "", "a", "test"]

    PIPE_PLUS: `"\|+"` Splits on one or more pipes. Ex:
              "this |is||a|test" = ["this ", "is", "a", "test"]
    """
    NOT_WORD = "\W+"
    WHITESPACE = "\s+"
    COMMA = ","
    COMMA_PLUS = ",+"
    PIPE = "\|"
    PIPE_PLUS = "\|+"


class Enrichment(ListableStrEnum):
    """This class contains constants for the names of enrichments"""
    AnomalyDetection = "anomaly_detection"
    Explainability = "explainability"



import pandas as pd
import numpy as np
import re

from arthurai.explainability.arthur_explainer import ArthurExplainer
from arthurai.common.constants import InputType, OutputType


def test_arthur_explainer_nlp_regression() -> None:
    def regression_predict(feature_vecs):
        results = []
        for fv in feature_vecs:
            results.append(np.array([0.2, 0.8]))
        return np.array(results)

    sample_data = pd.DataFrame([
        ['this is a test'],
        ['a second test with more words'],
        ['and a third test']
    ])

    explainer = ArthurExplainer(model_type=OutputType.Regression,
                                model_input_type=InputType.NLP,
                                num_predicted_attributes=1,
                                predict_func=regression_predict,
                                data=sample_data,
                                enable_shap=False,
                                label_mapping=[None])

    raw_feat_vecs = [
        ['first test test'],
        ['and a longer second test']
    ]
    exps = explainer.explain_nlp("lime", raw_feature_vectors=raw_feat_vecs, nsamples=100)

    # confirm we got 2 explanations
    assert len(exps) == 2

    for i, exp in enumerate(exps):
        # ensure single class
        assert len(exp) == 1
        # assert one explanation per unique word
        assert len(exp[0]) == len(re.split(explainer.text_delimiter, raw_feat_vecs[i][0]))


def test_arthur_explainer_nlp_classification() -> None:
    def regression_predict(feature_vecs):
        results = []
        for fv in feature_vecs:
            results.append(np.array([0.2, 0.7, 0.1]))
        return np.array(results)

    sample_data = pd.DataFrame([
        ['this is a test'],
        ['a second test with more words'],
        ['and a third test']
    ])

    explainer = ArthurExplainer(model_type=OutputType.Multiclass,
                                model_input_type=InputType.NLP,
                                num_predicted_attributes=3,
                                predict_func=regression_predict,
                                data=sample_data,
                                enable_shap=False,
                                label_mapping=[None])

    raw_feat_vecs = [
        ['first test test'],
        ['and a longer second test']
    ]
    exps = explainer.explain_nlp("lime", raw_feature_vectors=raw_feat_vecs, nsamples=100)

    # confirm we got 2 explanations
    assert len(exps) == 2

    for i, exp in enumerate(exps):
        # ensure single class
        assert len(exp) == 3
        # assert one explanation per unique word
        assert len(exp[0]) == len(re.split(explainer.text_delimiter, raw_feat_vecs[i][0]))



from typing import Any

import os

import pandas as pd
import responses
from arthurai.common.constants import InputType, OutputType

from arthurai import ArthurModel
from tests.base_test import BaseTest


class TestFileUpload(BaseTest):

    def _mock_file_upload(self, model_id, batch_id):
        server_response = {
            "counts": {
                "success": 100,
                "failure": 0,
                "total": 100
            },
            "failures": []
        }

        self.mockPost(f"/api/v3/models/{model_id}/inferences/file", server_response, status=207)
        self.mockPost(f"/api/v3/models/{model_id}/batches/{batch_id}", request_type=responses.PATCH, status=200,
                      dict_to_return={"message": "success"})

    def _mock_failed_batch_upload(self, model_id, batch_id):
        self.mockPost(f"/api/v3/models/{model_id}/inferences/file", {}, status=400)

    def _mock_failed_batch_close(self, model_id, batch_id):
        server_response = {
            "counts": {
                "success": 100,
                "failure": 0,
                "total": 100
            },
            "failures": []
        }

        self.mockPost(f"/api/v3/models/{model_id}/inferences/file", server_response, status=207)
        self.mockPost(f"/api/v3/models/{model_id}/batches/{batch_id}", request_type=responses.PATCH, status=404,
                      dict_to_return={"message": "batch not found"})

    @responses.activate
    def test_send_batch_inferences(self):
        model = self._set_up_test_and_get_model()
        resp, resp_close = model.send_batch_inferences(
            data=pd.DataFrame({
                "inference_timestamp": [1, 2, 3, 4],
                'partner_inference_id': ["1", "2", "3", "4"],
                "inference_timestamp": ['1', '2', '3', '4'],
                "partner_inference_id": ['1', '2', '3', '4'],
                "batch_id": ['1', '1', '1', '1']
            }),
            batch_id='batch_1')
        self._assert_result(resp)
        self.assertEqual(resp_close, {"dataset_close_result": "success"})

    @responses.activate
    def test_send_batch_inferences_auto_retry(self):
        model = self._set_up_test_and_get_model(failed_batch_upload=True)
        resp, resp_close = model.send_batch_inferences(
            data=pd.DataFrame({
                "inference_timestamp": [1, 2, 3, 4],
                'partner_inference_id': ["1", "2", "3", "4"],
                "inference_timestamp": ['1', '2', '3', '4'],
                "partner_inference_id": ['1', '2', '3', '4'],
                "batch_id": ['1', '1', '1', '1']
            }),
            batch_id='batch_1')
        self.assertEqual(len(responses.calls), 4)
        resp_400 = 0
        for http_call in responses.calls:
            if http_call.response.status_code == 400:
                resp_400 += 1
        self.assertEqual(resp_400, 4)

    @responses.activate
    def test_send_batch_close_auto_retry(self):
        model = self._set_up_test_and_get_model(failed_batch_close = True)
        resp, resp_close = model.send_batch_inferences(
            data=pd.DataFrame({
                "inference_timestamp": [1, 2, 3, 4],
                'partner_inference_id': ["1", "2", "3", "4"],
                "inference_timestamp": ['1', '2', '3', '4'],
                "partner_inference_id": ['1', '2', '3', '4'],
                "batch_id": ['1', '1', '1', '1']
            }),
            batch_id='batch_1')
        self.assertEqual(len(responses.calls), 5)
        resp_404 = 0
        for http_call in responses.calls:
            if http_call.response.status_code == 404:
                resp_404 += 1
        self.assertEqual(resp_404, 4)
        self.assertEqual(resp_close, {'dataset_close_result': {"message": "batch not found"}})

    @responses.activate
    def test_send_batch_ground_truth(self):
        model = self._set_up_test_and_get_model()
        resp = model.send_batch_ground_truths(data=pd.DataFrame({"col": [1, 2, 3, 4]}))
        self._assert_result(resp)

    @responses.activate
    def test_send_parquet_batch(self):
        model = self._set_up_test_and_get_model()
        file_loc = os.path.join(os.path.dirname(__file__), 'data', 'inference_batch_data')
        resp, close_resp = model.send_batch_inferences(batch_id="batch_1", directory_path=file_loc)
        self._assert_result(resp)
        self.assertEqual(close_resp, {"dataset_close_result": "success"})

    def _set_up_test_and_get_model(self, failed_batch_upload: bool = False, failed_batch_close: bool = False):
        model_data = {
            "partner_model_id": "",
            "input_type": InputType.Tabular,
            "output_type": OutputType.Regression,
            "display_name": "",
            "description": "",
            "attributes": [],
        }
        model = ArthurModel(client=self.arthur.client, **model_data)
        model.id = "1234"
        if failed_batch_upload:
            self._mock_failed_batch_upload("1234", "batch_1")
        elif failed_batch_close:
            self._mock_failed_batch_close("1234", "batch_1")
        else:
            self._mock_file_upload("1234", "batch_1")
        return model

    @staticmethod
    def _assert_result(resp: Any, success_count: int = 100, failure_count: int = 0, total_count: int = 100):
        assert resp["counts"]["success"] == success_count
        assert resp["counts"]["failure"] == failure_count
        assert resp["counts"]["total"] == total_count
